
const WEService        = require('../service');
const Models           = require('../models');
const WEConfig         = WEService.WEConfig;

module.exports = class OverseerService extends WEService {


  /**
   * Constructor.
   */
  constructor(options = {logger: WEConfig.logger}) {
    super({name : 'OverseerListener', port : WEConfig.get('WOV_we_ov_port'), logger: options.logger});

    this.RemoteListenerModel = null;
    this.ChannelModel = null;

    this.isrunning = false; // NOTE: set to null to turn it off
    this.activeLoop = false; // true when doWake handled data
  }


  /** */
  async onInit() {
    await super.onInit();

    this.RemoteListenerModel = Models.remoteListenerConnector(this.db);
    this.ChannelModel = Models.channelConnector(this.db);
  }


  /** */
  async onPostStartup() {
    await super.onPostStartup();
    if ( this.isrunning !== null ) {
      this.logger.info('... turning on doWake loop.');
      await this._startRunLoop();
    } else {
      this.logger.info('... doWake loop not being turned on.');
    }
  }

  /**
   * Starts the loop. Ensures one pass of doWake.
   * NOTE: no check on if already running!
   * NOTE: Previously manualStartLoop
   */
  async _startRunLoop() {
    this.isrunning = true;
    this.activeLoop = true;
    await this.doWake();
  }

  /**
   * Repeatedly call this.
   * @param {bool} _force - force run if true
   */
  async doWake(_force = false) {
    let didsomething = false;

    this.logger.h3().with('color', 'grey')
      .info(' *** doWake *****************************************************');

    await this.deleteNeverInitedChannels();

    await this.checkRemoteListeners();

    await this.recoverChannelsWithNoFunctioningRemoteListener();

    // TODO: this is not yet implemented
    await this.scaleRemoteListeners();

    this.activeLoop = didsomething;
    if ( didsomething ) setTimeout( this.doWake, 1000);
    else setTimeout( this.doWake, 5000);
  };


  /**
   * This would run a callback to determine if the Overseer needs to kill off or start up a new RemoteListener.
   * TODO: this is not yet implemented
  */
  async scaleRemoteListeners() {
    // this.logger.info(`... scaleRemoteListeners : not implemented`);
  }


  /**
   * Shouldn't happen, but it is a state that can occur. Push those channels back to 
   * inactive since nothing is listening to them. They will be managed again.
   */
  async recoverChannelsWithNoFunctioningRemoteListener() {

    // get all ids of remote listeners
    let rlidsresult = await this.RemoteListenerModel.find({}).select('_id'); // .map( (a) => { return a._id});
    let rlids = [];
    rlidsresult.forEach(function(v) {rlids.push(v._id)});
    if ( rlids.length > 0 ) {this.logger.info('rlids ', rlids);}

    // get all channels that don't have an id in rlids
    let channels = await this.ChannelModel.find({remotelistenerID : {$nin : rlids}, 'isActive' : {$in : ['running', 'starting']}});
    if ( channels.length > 0 ) {this.logger.info('channels: ', channels);}

    if ( channels.length > 0 ) {
      this.logger.info(`... recoverChannelsWithNoFunctiningRemoteListener: found ${channels.length} channels!?!`);
      for (let i=0; i<channels.length; i++ ) {
        let channel = channels[i];
        channel.isActive = 'inactive';
        channel.markModified('isActive');
        await channel.save();
      }
    }
 }


  /**
   * Turn off malfunctioning remote listeners.
   *
   * @param {int} _outofdateseconds - number of seconds to consider something out of date
  */
  async checkRemoteListeners(_outofdateseconds = -1) {
    let outofdate = (_outofdateseconds == -1 ? WEConfig.get('WOV_we_outofdateseconds') : _outofdateseconds); // seconds
    this.logger.info('out of date seconds ', outofdate);

    let from = new Date();
    from.setSeconds( from.getSeconds() - outofdate );
    let rls = await this.RemoteListenerModel.find( {updatedAt : { $lt : from}});

    this.logger.info('... checkRemoteListener: found RemoteListeners that are malfunctioning: # ',rls.length);
    for (let i=0; i<rls.length; i++ ) {
      let rl = rls[i];

      this.logger.info(`... checkRemoteListener: RemoteListener ${rl}`);
      // set isValid to false, so RL will shutdown immediately
      rl.isValid = false;
      rl.markModified('isValid');
      await rl.save();

      // mark each channel as inactive, so they will start to be picked up by other RLs
      let channels = await this.ChannelModel.find( {remotelistenerID: rl._id});
      this.logger.verbose(`... checkRemoteListener: has ${channels.length} channels`);
      for (let j=0; j<channels.length; j++ ) {
        let channel = channels[j];
        this.logger.verbose(`... checkRemoteListener: setting channel 'inactive' : ${channel}`);

        channel.isActive = 'inactive';
        channel.markModified('isActive');
        await channel.save();
      }

      // not sure what it would throw as an error... the promise returns 
      let result = await rl.remove();
      if ( result == null ) 
      { this.logger.throwError('Warning: Failed deleting RemoteListener after invalidating it: ', rl); }
    }
  };


  /**
   * Delete all channels that failed to get past initing, over 30 minutes ago by default.
   * @param {int} _outofdateseconds - time in seconds until a channel, never inited, is assumed to be fogotten, so it will be deleted
  */
  async deleteNeverInitedChannels(_outofdateseconds = (30*60)) {
    let outofdate = _outofdateseconds; // seconds
    let from = new Date();
    this.logger.info('from : ', from, outofdate);
    from.setMinutes( from.getMinutes() - outofdate );
    this.logger.info('from : ', from);
    let result = await this.ChannelModel.deleteMany( {isActive : 'initing', updatedAt : {$lt : from}});
    this.logger.info('result : ', result);
    if ( result.ok == 1 ) { // this was hit on overseer running in cluster
      // if ( result.result.n > 0 ) {this.logger.verbose('... deleteNeverInitedChannels: removed out of date initing channels: # ', result.result.n);}
      if ( result.n > 0 ) {this.logger.verbose('... deleteNeverInitedChannels: removed out of date initing channels: # ', result.n);}
    } else {this.logger.throwError(`Failed deleting 'initing' Channles, before updatedAt: ${from}`);}
  };

};
