

const WEService = require('../service');
const RemoteResHandler = require('./remoteResHandler');
const Long      = WEService.DB.Mongoose.Types.Long;


/**
 * This class has two functions:
 *   1) Sip older data from RemoteServices
 *   2) Connect and listen to RemoteServices for messages
 */
class RemoteDataSipper extends RemoteResHandler {

  /**
   *
   * @param {object} _options -
   */
  constructor(_options = {} ) {
    let options = Object.assign({}, {
      name         : 'RemoteDataSipper',
      port         : null,
      sipsize      : 10,
      FirstMsgTest : null,
      logger       : null,
    }, _options);
    // if ( _options.sipsize == null ) _options.sipsize=10;
    super(options); // {name, port, logger});
//    this._options.sipsize = sipsize;
//    this._options.FirstMsgTest = FirstMsgTest;

    if ( this._options.block_sip) {
      this.logger.h1().warn('Yo, set options.block_sip (above) to true, to block sipping data... only listening');
    }

    if ( this._options.sipsize == null ) {
      this.logger.throwError('the option "sipsize" is not set on remoteDataSipper, needed to tell how many posts to sip at a time');
    }
  }

  // ---------------------------------------------------------------------
  // Overrides
  // ---------------------------------------------------------------------
  //   Override these methods to implement a remotelistener service for
  //   a specific remote server/service.
  // ---------------------------------------------------------------------

  /**
   * Fetch messages from time _inc, up to _limit messages, on the given channel referenced by _channelref.
   *
   * @param {Long} _inc - Inc of the service.
   * @param {int} _limit - max messages to fetch (but could return fewer, and not always because there aren't more!)
   * @param {object} _channelref - The reference to channel infomation (cref and lref).
   * @return {object} : { success : bool, data : [] , curinc : Long }, where curinc is the last increment checked
  */
  async fetchOnAndBefore(_inc, _limit, _channelref) {
    try { throw new Error('You need to implement this for this remotelistener.'); }
    catch (err) { console.log(err); console.log('...continuing.'); }
    return {success : false, data : [], curinc : new Long(-1, 0)};
  };


  /**
   * "Current Time or Increment" of this service. NOT the remote service, since that would
   * be a guess at the RS's clock. This is basically used when starting the 'sipping'
   * of data from the remote service.
   * _channellocal will be found if it is null (some times channellocal has not been set yet)
   *
   * @param {object} _channel - channel Document
   * @param {object} _channellocal - channel local data structure
   * @return {long} - increment of the server. defaults to server time.
   */
  async getRSInc(_channel, _channellocal = null) {
    try { throw new Error('You need to implement this for this remotelistener.'); }
    catch (err) { console.log(err); console.log('...continuing.'); }
    // return WEService.l(new Date().getTime());
    return null;
  }

  // ---------------------------------------------------------------------
  // /Overrides
  // ---------------------------------------------------------------------


  /**
   * Sip data in a friendly way from the remote service for each channel. Forwards info to Woveon.
   *
   * @return {bool} - true if actively sipping information
  */
  async doSipDataFromChannels() {
    let retval = false;
    this.logger.aspect('sipChannelData', '***********************doSipDataFromChannels : start');

    if ( this._options.block_sip == true ) return false;

    let ckeys = Object.keys(this.myChannels);
    this.logger.aspect('sipChannelData', '  runLoop channel keys', ckeys);

    // for each channel, sip some data until first message found
    for (let i = 0; i<ckeys.length; i++) {
      let ctoken = ckeys[i];
      let channel = this.myChannels[ctoken];

      if ( channel.isActive != 'running' ) {
        this.logger.aspect('sipChannelData', `  ...skipping sip of channel : ${ctoken} : isActive: ${channel.isActive}`);
        break;
      }

      // if (channel.isActive && channel.e_first == null)
      // this.logger.aspect('sipChannelData', '  sipping on runLoop channel ', JSON.stringify(channel));
      try {
        let forwardretries = 1; // track number of sends
        let sipretries   = 2;
        let needstosip = false;   // did this do any sipping of data
        let hassipped = false;    // if it needed to sip, did it succeed?
        let hasforwarded = false; // if it sipped, did it successfully forward to Woveon?
        let data = null;

        // Try to sip data multiple times
        while ( sipretries > 0 && hassipped == false ) {

          // grab the data
          data = await this.sipChannelData(this._options.sipsize, channel);
          if ( data.data.length > 0 ) needstosip = true;
          this.logger.aspect('sipChannelData', 'sipped data : ', JSON.stringify(data, null, '  '));
          // FIXME: there is no error return value, so just assume it went ok
          hassipped = true;
          // if ( data.data.length != 0 ) { hassipped = true; }
          // else { this.logger.warn(`Failed sipping data ${ctoken}: `, channel); }

          sipretries--;
        }

        // if needs to sip and sipped worked, try forwarding the data
        while ( (needstosip == true) && (hassipped == true) && (hasforwarded == false) && (forwardretries > 0) ) {

          this.logger.aspect('sipChannelData', '******************* calling handleResForWoveon');
          let wovres = await this.convertChannelDataToRes(ctoken, data.data);
          let cr = await this.handleResForWoveon(ctoken, wovres);
          this.logger.aspect('sipChannelData', '******************* called handleResForWoveon, ', cr);
          if ( cr == 0 ) {
            this.logger.aspect('sipChannelData', 'sipChannel successfully forwarded Data to Woveon');
            hasforwarded = true;
          }
          else if ( cr && 2 ) {
            this.logger.error(`error, turn off channel ${ctoken} ${cr}: `, data.data, channel);
            this.logger.error('Failed to forward data to woveon. Is there an entry for this channel in apidb?');
            channel.isActive = 'error';
            channel.errorMessage='Failed to forward data to woveon. Is there an entry for this channel in apidb?';
            channel.markModified('isActive'); // what a PITA
            channel.markModified('errorMessage'); // what a PITA
            await channel.save();
            await this.doStopChannel(this.myChannels[channel.token], true); // true retains teh error state
            forwardretries = 0;
          }
          else if ( cr && 1 ) { this.logger.aspect('sipChannelData', 'sipChannel error... should retry.'); }
          else {
            this.logger.throwError(`Unhandled result of handleResForWoveon ${ctoken} ${cr} : `, data.data);
          }
          forwardretries--;
        }

        if ( needstosip ) retval = true; // if we need to sip, succeed or not, return true

      } catch (err) {console.log(err); throw err;}
    }
    this.logger.aspect('sipChannelData', '*******************doSipDataFromChannels : end');

    return retval;
  }



  /**
   * Called to go query up to _limit msgs from the remote service. 
   * 
   * This was made to be called repeatedly. So, some times it does not reutnr messages, but might next time. No worries,
   * we're sipp'n.
   *
   * This starts and can possibly create a new block or merge blocks.
   *
   * @param {*} _limit - number seeking to retrieve
   * @param {object} _channel - Mongoose Channel document (or one simulated with that data)
   * @return {array} - array of the Remote Service's messages that need to be processed (i.e. turn into Res and sent to Woveon)
   */
  async sipChannelData(_limit, _channel) {
    let retval = {data : [], pruned : [] };
    this.logger.aspect('sipChannelData', `**** start: sipChannelData :  limit: ${_limit}  channel: ${JSON.stringify(_channel)}.`);
    this.logger.aspect('sipMonitor', `**** start: sipChannelData :  limit: ${_limit}  channel: ${_channel.token}.`);
    let b0 = _channel.sipBlocks[0];
    // this.logger.aspect('sipChannelData', `...current sipBlock:  ${JSON.stringify(b0, null, '  ')}`);
    if ( b0 == null ) { this.logger.aspect('sipChannelData', `...no current sipBlocks`); }
    else if ( b0.finished == true ) { this.logger.aspect('sipChannelData', `...current sipBlock finished`); }
    else { this.logger.aspect('sipChannelData', `...current sipBlock:  ${b0}`); }

    this.logger.aspect('sipChannelData.quick', `**** start: sipChannelData :  limit: ${_limit}  channel.e_mon: ${JSON.stringify(_channel.e_mon)}   channel.sipBlocks : ${_channel.sipBlocks}.`);
    this.logger.aspect('sipChannelData.quick', `...current sipBlock:  ${b0}`);

    // sipBlock: F: first, P: pointer, L: latest
    if ( b0 == null ) {
      this.logger.aspect('sipChannelData', `  ... case 1`);
      let rsinc = await this.getRSInc(_channel); // choose 'now' which is kina arbitrary... following doLoop will fill in gaps using server times.
      this.logger.aspect('sipChannelData', '  ...getRSInc returned', rsinc);
      if ( rsinc.lessThan(_channel.e_mon) ) {
        this.logger.aspect('sipChannelData', `  ...since rsinc < _channel.e_mon (${_channel.e_mon}), set rsinc to that`);
        rsinc = _channel.e_mon;
      }
      b0 = {f: null, p: rsinc,  l: rsinc, finished: false, pid: null, lid: null};
      b0 = this.initFirstSipBlock(b0);
      this.logger.info('post initFirstSipBlock', b0, '... where f should not be null');
      _channel.sipBlocks.unshift(b0);
      this.logger.aspect('sipChannelData', '  ...after case 1, sipBLocks[0] ', _channel.sipBlocks[0]);
    }

    // spawn contiguous block to fill gaps caused by restart
    //  b0         b1   M
    //  F      P  LF    L
    //  [------***][----]
    else if ( b0.l.lessThan(_channel.e_mon) ) {
      this.logger.aspect('sipChannelData', `  ... case 2`);
      let b1 = { f: b0.l, p: _channel.e_mon, l: _channel.e_mon, finished : false, pid: null, lid: null };
      _channel.sipBlocks.unshift(b1);  // put at start of list so works on first
      this.logger.aspect('sipChannelData', '    ...after handling case 2 ', _channel, _channel.sipBlocks);
      b0 = b1;
    }

    // found start so merge blocks
    //  NOTE: b0.f == b0.p just means the inc are the same, but for time-based inc, there could be
    //        the rare case of two messages of exactly the same time, but the request only returned one,
    //        (edge case) so the b0.f == b0.p, but it is not finished.
    //  b1     b0
    //  [*****][****]
    else if ( b0.finished == true ) {
      this.logger.aspect('sipChannelData', `  ... case 3 : `, _channel.sipBlocks);
      let b1 = _channel.sipBlocks[1];
      if ( b1 != null ) { // merge blocks if we have more
        this.logger.info('b0:', b0);
        if ( ! b0.f.equals(b1.l) ) {
          this.logger.error('b0 (later)   : ', b0);
          this.logger.error('b1 (earlier) : ', b1);
          this.logger.throwError('non-contiguous blocks');
        }
        _channel.sipBlocks[1].l = b0.l; // expand block b1 to the right edge of b0
        _channel.sipBlocks[1].lid = b0.lid;
        _channel.sipBlocks.shift();     // pull off b0, so b1 is in front
      }
      b0 = b1;
    }

    // now, if block is not finished, fetch data back from b0.p, as much as limit allows
    //  NOTE: limit items are not guaranteed to be returned
    if ( _channel.sipBlocks[0].finished == false && _limit > 0 ) {

      this.logger.aspect('sipChannelData', `  ... block not finished`);

      // here... fetching back block limit... causing problem... so need to end at block limita (i.e. 'f')


      //TODO: fetchOnAndBefore should return Res objects
      let result = await this.fetchOnAndBefore(_channel.sipBlocks[0].p, _limit, _channel);

      if ( result == null || result.curinc === undefined || result.data == null ) {
        this.logger.error('result: ', result);
        this.logger.throwError('Badly formed result: ');
      }

      // call before prune since we are checking results, not what is pushed
      this.logger.aspect('sipChannelData', 'reached ? ', this.isSipBlockFinished(result, _channel, _limit) );

      if ( this.isSipBlockFinished(result, _channel, _limit) ) {
        _channel.sipBlocks[0].finished = true;
      }

//      this.logger.info('pre prune', result);
      result = this.pruneOnAndBeforeFetchedData(result, _channel); // remove duplicate on end
//      this.logger.info('post prune', result);


      // update channel based upon results, if we have them. Set p to the earliest result returned and l to last (furthest into future)
      if ( result.data.length != 0 ) {
        this.logger.aspect('sipChannelData', '  ... setting sip.p and sip.l id since we have messages');
        _channel.sipBlocks[0].pid = this.genMsgRemoteID(result.data[0]);   // update the id of that data
        if ( _channel.sipBlocks[0].lid == null ) { // update lid if we have data and it isn't set
          _channel.sipBlocks[0].lid = this.genMsgRemoteID(result.data[result.data.length-1]);
        }
      } else {
        if ( _channel.sipBlocks[0].p != result.curinc ) {
          this.logger.aspect('sipChannelData', '  ... setting p/l id null, since we have no messages but did move Inc ');
          _channel.sipBlocks[0].pid = null;
        }
      }
      // update pointer based on trimmed result data
      _channel.sipBlocks[0].p = result.curinc;
      this.logger.info('NOT Sure if this is a real note: here, should set l if p > l');

      this.logger.aspect('sipChannelData', ' channel sipblocks after fetch: ', _channel.sipBlocks);

      // due to testing, not all channels have a 'save'
      if ( _channel.save != null ) {
        this.logger.aspect('sipChannelData', '  ...saving channel to database');
        _channel.markModified('sipBlocks');
        await _channel.save();
      }

      retval = {data : result.data || [], pruned : result.pruned || [] }; // may be [], but not null
    }

    this.logger.aspect('sipChannelData', `  ... done. retval : ${JSON.stringify(retval)}`);
    return retval;
  }


  /**
   * Basically, a notification of new messages just came in. So, add a sipBlock up front, to sip those.
   *
   * If no sipblocks, create one. Else create a new one that abuts the current.
   *
   * Update e_mon to this new time
   */
  async sipFromFront(_newLInc, _channel) {
    let b0 = _channel.sipBlocks[0];
    this.logger.aspect('sipFromFront', 'sipFromFront: ',_newLInc, ' : ', _channel);
    this.logger.aspect('sipFromFront', `  ...sipBlocks:  `, _channel.sipBlocks);

    // no blocks so create first
    if ( b0 == null ) {
      this.logger.aspect('sipFromFront', `  ...sipFromFront case 1 : no block so create first`);
      b0 = {
        f: null, 
        p: Long.fromBits(_newLInc.getLowBitsUnsigned(), _newLInc.getHighBits()), 
        l: Long.fromBits(_newLInc.getLowBitsUnsigned(), _newLInc.getHighBits()), 
        finished: false, 
        pid: null, 
        lid: null
      };
      b0 = this.initFirstSipBlock(b0);
      _channel.e_mon = Long.fromBits(_newLInc.getLowBitsUnsigned(), _newLInc.getHighBits());
      this.logger.aspect('sipFromFront', 'post initFirstSipBlock', b0, '... where f should not be null');
      _channel.sipBlocks.unshift(b0);
    } 

    // take a look at existing block
    else {
      this.logger.aspect('sipFromFront', `  ...sipFromFront case 2 : look at existing blocks`);

      // if there is a change, update the sipblocks
      if ( b0.l.lessThan(_newLInc) ) {
        // ermehgesh Becky, just create a new block why don't you
        this.logger.aspect('sipFromFront', `  ...sipFromFront case 3: b0.l is less than newLInc`);
        let b1 = { 
          f: Long.fromBits(b0.l.getLowBitsUnsigned(), b0.l.getHighBits()), 
          p: Long.fromBits(_newLInc.getLowBitsUnsigned(), _newLInc.getHighBits()), 
          l: Long.fromBits(_newLInc.getLowBitsUnsigned(), _newLInc.getHighBits()), 
          finished : false, 
          pid: null, 
          lid: null 
        };
        _channel.e_mon = Long.fromBits(_newLInc.getLowBitsUnsigned(), _newLInc.getHighBits());
        _channel.sipBlocks.unshift(b1);
      } else { 
        this.logger.aspect('sipFromFront', `  ...sipFromFront case 4 : b0.l is greater than or equal to _newLInc`);
      }
    }
    this.logger.aspect('sipFromFront', `  ...sipFromFront 5 : now update channel e_mon`);

    this.logger.aspect('sipFromFront', '  ...sipFromFront: ',_newLInc, ' : ', _channel, ' : ', _channel.sipBlocks);

    // Save block data
    // NOTE: due to testing, not all channels have a 'save'
    if ( _channel.save != null ) {
      this.logger.aspect('sipFromFront', `  ...sipFromFront case 7`);
      this.logger.aspect('sipFromFront', '  ...saving channel to database');
      _channel.markModified('e_mon');
      _channel.markModified('sipBlocks');
      await _channel.save();
    }
    this.logger.aspect('sipFromFront', `  ...sipFromFront case 8`);
  }


  /**
   * Remove last results, if it has already been returned.
   *
   * This handles cases when remote services can return a message on the exact time, so
   * a query would have to be inclusive of the time of the last message returned, resulting
   * in a duplicate.
   *
   * NOTE: I'm assuming they will return in the same order each time. I.e. all same date/time
   *       nut remote service always returns them in a certain way.
   * @param {Array} _result - messages fetched from the remote server.
   * @param {object} _channel - The channel object.
   * @return {object} - _result, with msgs modified
   */
  pruneOnAndBeforeFetchedData(_result, _channel) {
    let i = _result.data.length-1;
    this.logger.aspect(`pruneOnAndBefore`, 'result : ', _result, ' sip :', _channel.sipBlocks);

    // no data, no prune
    if ( _result.data.length == 0 ) { }

    // if no pointing id, then no pruning
    else if ( _channel.sipBlocks[0] == null || _channel.sipBlocks[0].pid == null ) {}

    // start at end, and while same time, look for postIDs, finding one, return earliest
    else if ( _channel.sipBlocks[0].pid != null ) {
      let pruned = [];
      this.logger.aspect(`pruneOnAndBefore`, `i: ${i}  i-th msg`+
                         ` t:${this.genMsgInc(_result.data[i], _channel)}`+
                         ` earliest t:${_channel.e_earliest}`);

      let cont = true;
      while ( cont==true && this.genMsgInc(_result.data[i], _channel) == _channel.sipBlocks[0].p ) {
        this.logger.aspect('pruneOnAndBefore', `  - compare pos ${i} : ${this.genMsgRemoteID(_result.data[i])}, to earliestID: ${_channel.sipBlocks[0].pid}`);
        if ( this.genMsgRemoteID(_result.data[i]) == _channel.sipBlocks[0].pid) { cont = false; }
        let p = _result.data.splice(i, 1);
        pruned.unshift(p[0]);
        i--;
        if ( i < 0 ) cont = false;
      }
      _result.pruned = pruned;
    }

    this.logger.info('  resulting pruned: ', _result.pruned);

    return _result;
  }


  /**
   * TLDR; Override this function to change assumptions about when to assume the first message for a channel has been rached.
   * 
   * Details: Most time-based RemoteServices you can assume will send their last message through normal requests. I.e.
   * the request is effectively, send previous messages before time X. However, some (IMAP) are not, they say
   * "send me range x-y", so you have to limit the range size. In those cased, no messages could be returned, but
   * there could be messages less than X. 
   * 
   * @param {*} _result  - result of fetch
   * @param {*} _channel - channel information, updated for the _result
	 * @param {int} _limit - Number of messages requested
   */
  isFirstMsgReached(_result, _channel, _limit) {
    let retval = false;

    if ( this._options.FirstMsgTest === undefined) {
      this.logger.throwError("You need to provide a 'FirstMsgTest' option for this RemoteDataSipper, or override 'isFirstMsgReached'.");
    }

    if ( this._options.FirstMsgTest == RemoteDataSipper.FMT_ResultLengthZero ) {
      if ( _result.data.length == 0 ) retval = true;
    } 
    if ( retval == false && this._options.FirstMsgTest == RemoteDataSipper.FMT_LimitNotReached ) {
      if ( _result.data.length != _limit  ) retval = true;
    } 
    return retval;
  }


  /**
   * Call on each incoming message from Remote Service (i.e. webhook, imap, etc.). This
   * compares the message time to the channel to see if we need to forward to API. Updates
   * channel as well on success.
   *
   * @param {Channel} _channel - channel object
   * @param {string} _msg - server time of this message (i.e. the (U)pdate time of the message)
   * @return {boolean} True if should forward (and indicates channel was modified)
   */
  updateSipperWithNotification(_channel, _msg) {
    let retval = false; // always start false

    this.logger.aspect('sippernotification', `genMsgInc with msg :`, _msg);
    this.logger.aspect('sippernotification', `  on channel: `, _channel);
    let inc    = this.genMsgInc(_msg, _channel); // ??? is this local channle object?
    this.logger.aspect('sippernotification', 'inc :', inc);

    this.logger.aspect('sippernotification', `isValidRSNotification: t:${JSON.stringify(inc)} : ${JSON.stringify(_channel)} : msg: ${JSON.stringify(_msg)}`);
    this.logger.aspect('sippernotification', `sipBlocks:`, _channel.sipBlocks);
    // no blocks yet, so create
    if ( _channel.sipBlocks == null || _channel.sipBlocks.length == 0 ) {
      this.logger.aspect('sippernotification', `case 1 `);
      let b = {f : null, p : inc, l : inc, finished : false, pid : this.genMsgRemoteID(_msg),  lid : this.genMsgRemoteID(_msg)};
      b = this.initFirstSipBlock(b);
      _channel.sipBlocks.unshift(b);  //
      retval = true;

    } else {

      this.logger.aspect('sippernotification', `case 2 `, _channel.sipBlocks);

      // ignore message below pointer, will get to it with sipChannelData (not going to represent non-contiguous posts)
      // NOTE: could be in other blocks, but ignoring
      if (inc.lessThanOrEqual(_channel.sipBlocks[0].p)) {
        this.logger.aspect('sippernotification', `case 2.1 `);

      // ignore below last, should be in what has already been sipped
      } else if ( inc.lessThan(_channel.sipBlocks[0].l) ) {
        this.logger.aspect('sippernotification', `case 2.2 `);

      // add after last, since assuming msgs coming in, in real time
      } else if ( inc.greaterThan(_channel.sipBlocks[0].l) ) {
        this.logger.aspect('sippernotification', `case 2.3  ${WEService.i(inc)} ${WEService.i(_channel.sipBlocks[0].l)}`);

        // been monitoring so this is a message to keep
        _channel.sipBlocks[0].l = inc;
        _channel.sipBlocks[0].lid = this.genMsgRemoteID(_msg);
        retval = true;

      // equal to current last? investigate
      } else if ( inc.equals(_channel.sipBlocks[0].l) ) {
        this.logger.aspect('sippernotification', `case 2.4 `);

        // check ids... since Inc not always unique (i.e. FB uses time in seconds), so this accepts border cases, after checking Id
        // POTENTIAL BUG: if three come in with same Inc, then just checking Ids will not suffice since they can be in different orders (hopefully the remote service, even with same Incs, resends each time in the same order to avoid this rare bug!).
        let msgid = this.genMsgRemoteID(_msg);
        this.logger.info('  compare ids :', msgid, _channel.sipBlocks[0].lid);
        if ( msgid != _channel.sipBlocks[0].lid ) {
          this.logger.aspect('sippernotification', `case 2.4.1: same Inc, different ids, so keep `);
          // been monitoring so this is a message to keep
          _channel.sipBlocks[0].l = inc;
          _channel.sipBlocks[0].lid = this.genMsgRemoteID(_msg);
          retval = true;
        } else {
          this.logger.aspect('sippernotification', `case 2.4.2: same Inc, same id so skip `);
        }

      } else {
        this.logger.throwError(`How did you get here? time:${inc}`);
      }
    }

    this.logger.aspect('sippernotification', `  ... on end sipBlocks:`, _channel.sipBlocks);
    this.logger.aspect('sippernotification', `  ... returning ${retval} `);
    return retval;
  }


  /**
   * Determines if the current SipBlock has completed
   */
  isSipBlockFinished(_result, _channel, _limit) {
    let retval = false;
    this.logger.info('isSipBlockFinished :', _result);
    if ( _result.curinc == null || _result.curinc.equals(_channel.sipBlocks[0].f) ) retval = true;
    this.logger.aspect('sipChannelData', `  ... isSipBlockFinished : ${_result.curinc} == ${_channel.sipBlocks[0].f} ? ${retval}`);
    return retval;
  }

  /**
   * Call on 1st message. Assumes '0', but other plugins might have different 
   * start message.
   * @param {object} _sipblock - initializes this block
   * @return {object} - returns the sipblock
   */
  initFirstSipBlock(_sipblock) {
    _sipblock.f =  WEService.l(0);
    this.logger.aspect('sipChannelData', 'initFirstSipBock : ', _sipblock);
    return _sipblock;
  }

};


/** Options for FirstMsgTest */              // Found first message when...
RemoteDataSipper.FMT_ResultLengthZero = 1;   //   "When no messages are returned"
RemoteDataSipper.FMT_LimitNotReached  = 2;   //   "When fewer messages returned than what we asked for"

module.exports = RemoteDataSipper;


