
const moment           = require('moment');
const WovReturn        = require('woveon-service').WovReturn;
// const ReqWoveon        = require('../service/reqWoveon');
const Models           = require('../models');
const RemoteDataSipper = require('./remoteDataSipper');
// const MongoDb          = require('mongodb');


/**
 * This has two functions:
 *   1) Listen to WoveonListener to start new channels.
 *   2) Wake and check still alive, and grab new channels if possible.
 */

module.exports = class RemoteListenerService extends RemoteDataSipper {

  /**
   * Constructor.
   * @param {object} _options
   */
  constructor(_options = {}) {
    let options = Object.assign({}, {
      name      : 'RemoteListener',
      baseroute : `/${RemoteDataSipper.WEConfig.get('WOV_we_ver')}`,
      port      : RemoteDataSipper.WEConfig.get('WOV_we_rl_port'),
      logger    : RemoteDataSipper.WEConfig.logger,
    }, _options);
    // if ( ! _options.requestip ) _options.requestip = true;
    super(options);

    this.toWoveon = null;            // moved this to onInit so the db is active

    // set in init()
    this.RemoteListenerModel = null; // was remotelistenerModel
    this.ChannelModel = null;

    // set in postStartup()
    this.remotelistenerEntry = null;
    this.myChannels          = {}; // indexed by channel token, These are Mongoose objects
    this.myChannelsLocalData = {}; // - indexed by channel token, These are json objects, not stored in
                                   //   database, persistent so long as this service is
      // this is probably a bad idea to store, but since we open up remote connections, then
      // these Services are actually stateful, so this stores that.

    this.isrunning = false;  // NOTE: set to null to turn it off. starts in false
    this.activeLoop = false; // true when doWake handled data
  }

  /**
   * Override for your own Service
   *
   * @return {fraction} - a fraction of 1, where 1 is 100%
   */
  getCapacity() { return 0; };


  /**
   * initialize this service and routes.
   */
  async onInit() {
    await super.onInit();

    // Create models
    this.RemoteListenerModel = Models.remoteListenerConnector(this.db);
    this.ChannelModel        = Models.channelConnector(this.db);
//    this.IncompleteRes       = Models.incompleteResConnector(this.db);
    // this.logger.info('MessageModel.connect', this.MessageModel);
//    await this.MessageModel.connect(this.db);


    // Requester to Woveon API
    // this.toWoveon            = new ReqWoveon(this.logger);

    /**
     * Complete the channel by registering with remote remotelistener
     * and start downloading data.
     *
     * @param {string} token - the channel's identifying token
     *
     * takes: { token : {string} }
     * returns: { success: true, data: true }
     */
    this.listener.onPost('/start_channel', this.doStartChannel, __filename);
    // TODO: Umm, delete these? start_channel / stop_channel since WL does not call this anymore

    /**
     * Deregister with remote remotelistener then get rid of channel.
     */
    this.listener.onGet('/stop_channel', this.doStopChannel, __filename);

    /**
     * Return healthy.
     */
    this.listener.onGet('/capacity', this.doCapacity, __filename);
  };


  /**
   * Sets this up once Listener is operating.
   */
  async onPostStartup() {
    await super.onPostStartup();

    // Grab from config (which is set by Listener.run())
    if ( this.listener.islistening == false ) {
      throw new Error('remotelistener does not have IP credentials yet, setup in Listener.run().');
    }

    this.remotelistenerEntry = new this.RemoteListenerModel({
      numchannels : 0,
      isValid     : true,
      capacity    : 0.0,

      // address used by WL
      address_ip     : this.listener.server.address().address,
      address_family : this.listener.server.address().family,
      address_port   : this.listener.server.address().port,
    });
    this.remotelistenerEntry.markModified('object'); // has to do on edits or else won't save
    let r = await this.remotelistenerEntry.save();
    this.logger.info('remotelistenersEntry save result : ', r);

    this.logger.info('Created remotelistener: ', JSON.stringify(this.remotelistenerEntry));
    if ( this.isrunning !== null ) {
      this.logger.verbose('... turning on doWake loop (and not blocking).');
      this._startRunLoop(); // NOTE: is not blocking!
    }
    else { this.logger.verbose('... doWake loop not being turned on.'); }
  };


  /**
   * Starts the loop. Ensures one pass of doWake.
   * NOTE: no check on if already running!
   * NOTE: Previously manualStartLoop
   */
  async _startRunLoop() {
    this.isrunning = true;
    this.activeLoop = true;

    // infinite loop (until externally set isrunning to false)
    while (this.isrunning) {
      await this.doWake();

      // delay based upon how much activity
      if ( this.isrunning ) {
        let timeoutms = 5000;
        if ( this.activeLoop ) timeoutms = 1000;
        await new Promise( (resolve) => setTimeout(resolve, timeoutms));
      }
    }
  }


  /**
   * Repeatedly call this.
   * @param {bool} _force - force run if true
   */
  async doWake(_force = false) {
    let didsomething = false;

    if ( this.isrunning == false && _force == false ) return; // break out of loop

    // Alive Checkin - do even if not 'running'
    try {
      if ( _force == true ) {
        this.logger.verbose('  ... skipping doAliveCheck due to _force');
      }
      else { await this.doAliveCheckin(); }
    }
    catch (e) { this.logger.error(e); process.exit(1); }

    this.logger.h3().with('color', 'grey')
      .info(' *** doWake *****************************************************');

    // Get own channels that aren't active
    if ( await this.doCheckToEnableOwnChannels() ) { didsomething = true; }

    // this.logger.info('*** turning off wake features for development');
    // Sip data from Channels that are active
    if ( await this.doSipDataFromChannels() ) { didsomething = true; }

    // Check to add channels if we have capacity
    if ( this.hasCapacity() ) {
      if ( await this.doSeekNewChannels() ) { didsomething = true; }
    }

    // configure next wake loop call
    this.activeLoop = didsomething;

    return didsomething;
  }


  /**
   * Tell the watcher that this RL is still alive, so don't take away its
   * channels. Assumes that if this RL can check in, that it can listen to
   * channels.
  */
  async doAliveCheckin() {

    // get own entry
    let rlentry = await this.RemoteListenerModel
      .findOne({_id : this.remotelistenerEntry._id})
      .exec();
    this.logger.aspect('alivecheck', `alive check for ${this.remotelistenerEntry._id} : `, rlentry);

    // if null or invalid, die. Basically, the Watcher is hard-killing this. this
    // process could still hang in its own loop, never reading the database do never exiting...
    // TODO: In watcher, instead of deleteing hte remoteListener row in db, move to an
    //       entry to track when it was 'killed' and then force delete the pod if it never
    //       self collects.
    try {
      if ( rlentry == null ) { throw new Error('No entry for this remoteListener anymore. Immediate exit!'); }
      if ( rlentry.isValid == false ) { throw new Error('Entry is invalid for this remoteListener. Immediate exit!'); }
    }
    catch (e) { this.logger.throwError(e); };

    // Update values (updateAt will auto-update).
    rlentry.set('capacity', this.getCapacity());
    rlentry.set('n', this.myChannels.length);
    rlentry.markModified('object'); // has to do on edits or else won't save
    try {
      this.logger.aspect('alivecheck', `  ...alive check save `, rlentry);
      await rlentry.save();
    }
    catch (e) { this.logger.throwError(e); }
  }


  /**
   * Determines if this should seek more channels.
   * @return {bool} - true if it thinks it can handle more channels.
  */
  hasCapacity() {
    if ( this.getCapacity() < 1.0 ) return true;
    return false;
  }


  /**
   * Get own channels that are not activated yet.
   * TODO: optimization: allow this method to opperate in parallel
   * @return {bool} - true if did something
  */
  async doCheckToEnableOwnChannels() {

    let numtohandle = 3;
    let retval = false;

    // get all
    let channels = await this.ChannelModel.find({isActive : 'starting', remotelistenerID : this.remotelistenerEntry._id, parentchannel : null});
    this.logger.aspect('doCheckToEnableOwnChannels', `...found ${channels.length} channels of mine that need to be enabled for rl ${this.remotelistenerEntry._id}`);
    if ( channels.length > 0 ) retval = true; // even if don't start something, there are things to start

    // pick 'numtohandle' random entries (i.e. don't deadlock on retrying first few every time!)
    for (let i=0; i<Math.min(numtohandle, channels.length); i++) {
      let pos = Math.floor(Math.random() * Math.floor(channels.length));
      this.logger.aspect('doCheckToEnableOwnChannels', `Looking at restarting my own channel {pos: ${pos}}, entry :`, channels[pos]);

      // recreate args from channel entry
      let args = channels[pos].pluginData;
      args.token = channels[pos].token;
      try {
        let r = await this._startOwnChannel(channels[pos], true, args);
        if ( r !== true ) { this.logger.throwError(r); }
      }
      catch (e) {
        this.logger.error('In doCheckToEnableOwnChannels, failed to start a channel it just found. This is either an intermitent connection error, or the server does not have knowledge of the channel and becomes a problem.');
      }
    }

    return retval;
  }


  /**
   * @return {bool} - true if tried to seek a new channel
  */
  async doSeekNewChannels() {

    let numtohandle = 3;
    let retval = false;

    // This grabs a bunch up front
    let channels = await this.ChannelModel.find({isActive : 'inactive', parentchannel : null}).limit(numtohandle);
    if ( channels.length > 0 ) retval = true; // even if don't start something, there are things to start

    this.logger.info(`doSeekNewChannels found inactive channels: `, channels);

 //   this.logger.h3().info('update to this RL id of ', this.remotelistenerEntry._id);
    this.logger.info(`doSeekNewChannels in ${channels.length} channels`);
    for (let i=0; i<channels.length; i++) {

      this.logger.info(`doSeekNewChannels : ${i}: ${channels[i].token}  - channel RL is `, channels[i].isActive);

      // do update and get, so we can 'collect' it atomically from other RemoteListeners. null if other grabbed it first
      let channelresult = await this.ChannelModel.findOneAndUpdate(
            {isActive : 'inactive', _id : channels[i]._id},
            {remotelistenerID : this.remotelistenerEntry._id},
            {new : true, rawResult : true});

      this.logger.info(`doSeekNewChannels :          ${i}  - channel RL find and update result: `, channelresult);
      if ( channelresult && channelresult.ok== 1 && channelresult.lastErrorObject.updatedExisting == true ) {

        let channel = channelresult.value;
        try {
          if ( channel != null ) {
            // start it if something was found and returned
            channels[i].remotelistenerID = channel.remotelistenerID; // set channels object, since channel is not a MongoooseScheme
            let r = await this._startOwnChannel(channels[i], true, null);
            if ( r !== true ) { this.logger.throwError(r); }
          }
        }
        catch (e) {
          this.logger.error('In doSeekNewChannels, failed to start a channel. Probably an intermitent connection error. Because we have claimed it, we will keep retrying.');
          this.logger.error(e);
        }
      }
    }

    return retval;
  }


  /**
   * This abstracts the process of signing up a channel for a service. A
   * service should extend/override these methods.
   * TODO: Question: what if this claims the channel, but then never activates it with Woveon... or fails
   *        with woveon... could lead to problems.
   * @param {object} _args - Directly passed from WoveonListener. Only 'token',
   *                         channel token identifying the channel with Woveon, will
   *                         be guaranteed.
   * @return {bool}
   */
  async doStartChannel(_args) {
    let retval = null;

    this.logger.verbose(`doStartChannel : ${JSON.stringify(_args)}`);

    // Find the channel
    let channel = null;
    try {

      channel = await this.ChannelModel.findOne({token : _args.token});
      if (channel == null) { throw new Error('Invalid channel token:', _args.token); }
      this.logger.info('channel is :', channel);
      if ( channel.remotelistenerID != null ) {
        throw new Error('Channel already taken by a remotelistener, so skipping: ', channel);
      }
    }
    catch (err) {
      this.logger.error(err); throw new Error(`Failed get/find channel with ${_args.token}.`);
    }

    // Grab the channel by writing RemoteListenerID to it. Do this before the connection to remote services.
    try {
      channel.remotelistenerID = this.remotelistenerEntry._id;
      await channel.save();
    }
    catch (err) {
      this.logger.error(err);
      throw new Error(`Failed writing remotelistenerID of `+
                      `'${JSON.stringify(this.remotelistenerEntry)}' `+
                      `to channel of ${_channel.token}.`);
    }

    let result = await this._startOwnChannel(channel, false, _args );
    this.logger.info('doStartChannel result ', result);
    if ( result == true ) { retval = WovReturn.retSuccess(result); }
    else { retval = WovReturn.retError(result, 'Failed to start a channel'); }

    return retval;
  }


  /**
   * Override as needed. This is the local data store in a RL. It is called by _startOwnChannel when
   * it is starting up.
   * ex. PLEM stores the imap datastructure and box data here.
   * @return {object} -
   */
  initChannelLocalData() { return {}; }


  /**
   * Internal class that starts a channel for /start_channel as well as seeking new channels and restarting your own
   * @param {object} _channel - channel object retrieved
   * @param {boolean} _doDisableRSConnection - if true, then removes connection to remote services before starting it
   * @param {*} _args -
   * @return {bool} - true on success, everything else is a failure
   */
  async _startOwnChannel(_channel, _doDisableRSConnection, _args) {

    this.logger.aspect('startownchannel', '_startOwnChannel called for :', _channel.token);

    // Grab the channel by writing RemoteListenerID to it. Do this before the connection to remote services.
    try {

      // add to this plem's channels
      this.myChannels[_channel.token] = _channel;
      _channel.isActive = 'starting';
      await _channel.save();
    }
    catch (err) {
      this.logger.error(err);
      throw new Error(`Failed writing remotelistenerID of `+
                      `'${JSON.stringify(this.remotelistenerEntry)}' `+
                      `to channel of ${_channel.token}.`);
    }


    // Turn off the connection to the RemoteService if we need to.
    try {
      if ( _doDisableRSConnection ) {
        let localchanneldata = this.myChannelsLocalData[_channel.token];
        this.logger.info('calling doDisableChannel 1');
        let result = await this.doDisableChannel({cref : _channel, lref : localchanneldata});
        this.logger.verbose('doDisableRSConnection result : ', result);
        // NOTE: result doesn't matter since we don't know if the connection is actually valid. If it errors,
        //       the catch should break this out. Normal success/fail will just continue normally, assuming that
        //       the RemoteService has no connection.
      }
    }
    catch (err) {
      this.logger.error(err);
      this.logger.throwError('Failed to do disable remote connection. Will retry next cycle.');
    }


    let channelsetup = true; // set to something other than true on failure
    // NOTE: i broke up the doConnectChanenlToRS from the toWoveon.activateChannel, so that this code is with teh claiming of the channel, by setting remotelistenerID
    // Setup the channel and subchannels
    try {

      // Configure something that might return data to this service, like webhooks.
      // NOTE that this is a ref to channel, so it can be modified.
      let re = this.myChannelsLocalData[_channel.token];
      if (re == null ) {
        this.myChannelsLocalData[_channel.token] = this.initChannelLocalData();
        re = this.myChannelsLocalData[_channel.token];
      }

      this.logger.aspect('startownchannel', '  ... calling doConnectChannel');
      let rsresult = await this.doConnectChannel({cref : _channel, args : _args, lref : re});
      if ( ! Array.isArray(rsresult) ) {
        this.logger.error('doConnectChannel failed: ', rsresult);
        // FIXME don't return early... set up a series of checks on retval, so it ends at the bottom
        return false;
      }
      if ( _channel.e_mon == null ) { this.logger.throwError('did not set e_mon. has to be set when talking with remote server.'); }
      this.myChannelsLocalData[_channel.token] = re; // update

      // Handle the subchannels that are created above
      this.logger.info('Handle the subchannels that are created above');
      for (let i=0; i< rsresult.length; i++ ) {
        let sc = rsresult[i];
        if ( sc.cref.e_mon == null ) { this.logger.throwError('did not set e_mon of subchannel: ', sc); }
        this.logger.info('add in new channel : ', sc.cref.token, Object.keys(sc));
        this.myChannels[sc.cref.token]          = sc.cref;
        this.myChannelsLocalData[sc.cref.token] = sc.lref;

        sc.cref.remotelistenerID = this.remotelistenerEntry._id; // reset since the doDisableChannel could remove it
        sc.cref.markModified('remotelistenerID'); // what a PITA, do this so it saves
        sc.cref.markModified('pluginData');       // what a PITA, do this so it saves
        sc.cref.isActive = 'running';
        sc.cref.markModified('isActive');         // what a PITA, do this so it saves
        sc.cref.errorMessage = '';                // reset errorMessage
        sc.cref.markModified('errorMessage');
        await sc.cref.save();
      }

      this.logger.aspect('startownchannel', 'myChannels :', JSON.stringify(this.myChannels[_channel.token]));
      this.logger.aspect('startownchannel', '  _channel :', JSON.stringify(_channel));

      if ( rsresult instanceof WovReturn ) { retval = rsresult; }
      if ( rsresult == null ) throw new Error('general doConnectChannel error');

      _channel.remotelistenerID = this.remotelistenerEntry._id; // reset since the doDisableChannel could remove it
      _channel.markModified('remotelistenerID'); // what a PITA
      _channel.markModified('pluginData'); // what a PITA
        // notice I am not writing back yet. no need to.
    }
    catch (err) {
      console.log(err);
      this.logger.throwError(`failed call to doConnectChannel for: ${_channel.token}.`, err);
      channelsetup = err.message;
      this.logger.info('channelsetup ', JSON.stringify(channelsetup));
    }

    // notify Woveon of channel activation if all above passes.
    /*
    ** moved into part of /remote_launch
    if ( channelsetup === true ) {
      try {

        // Post back to Woveon that the channel is active, and what its icon should be.
        // this.logger.h3().info('activate channel with credentails: ', this.toWoveon.credentials);
        let result = await this.toWoveon.activateChannel(_channel.token, _channel.pageImage);
        this.logger.info(`toWoveon.activateChannel returned : ${JSON.stringify(result)}`);
        if ( result.success == false ) {
          throw new Error(`toWoveon.activateChannel ${_channel.token} failed.`);
        }
      } catch (err) {
        this.logger.error(`failed call to activateChannel: ${_channel.token}.`, err);
        channelsetup = err;
      }
    }
    */

    // handle return value and cleanup
    if ( channelsetup === true ) {

      // Update remotelistener counter
      this.remotelistenerEntry.numchannels++;
      await this.remotelistenerEntry.save(); // TODO: necessary? the callback should handle (it doesn't have to be perfectly accurate!)
      _channel.isActive = 'running';
      _channel.errorMessage = '';                // reset errorMessage
      _channel.markModified('errorMessage');
    }
    else {
      this.logger.warn('Either auth error or token not on server', channelsetup);
      this.logger.warn('Marking channel isActive as "error".');
      _channel.isActive = 'error';
      _channel.errorMessage = ''+moment().format()+'] Either auth error or token not on server';
      _channel.markModified('errorMessage'); // what a PITA
    }

    // Save channel changes
    _channel.markModified('isActive'); // what a PITA
    let r = await _channel.save();
    this.logger.aspect('startownchannel', 'activated channel save result: ', JSON.stringify(r, null, '  '));

    // save changes back to myChannels
    this.logger.warn('I moved this code up, to save the channel to myChannels, so I need '+
                     'to ensure that the entry in myChannels for the channel, is the same for '+
                     '_channel after all the changes (i.e. pointer issue).');
    this.logger.info('myChannels :', JSON.stringify(this.myChannels[_channel.token]));
    this.logger.info('  _channel :', JSON.stringify(_channel));

    // stop channel if not running (and retain the isActive state (error))
    if ( _channel.isActive != 'running' ) {
      await this.doStopChannel(this.myChannels[_channel.token], true);
    }

    this.logger.aspect('startownchannel', 'returning channlesetup ', channelsetup);
    return channelsetup;
  };


  /**
   * Stop the channel and retain data. Updates the channel document, turns off remote
   * service and deactivates channel on Woveon.
   * @param {*} _args
   * @param {boolean} _retainIsActive -
   * @return {bool}
   */
  async doStopChannel(_args, _retainIsActive = false) {

    this.logger.verbose(`doStopChannel : ${this != null} : ${JSON.stringify(_args)}`);

    // Find the channel
    let channel = null;
    let localchanneldata = null;
    try {

      this.logger.info('All channel tokens: ', Object.keys(this.myChannels));

      channel = this.myChannels[_args.token];// await this.findChannel(_args.token);
      if (channel == null) { throw new Error('Invalid channel token:', _args.token); }

    }
    catch (err) {
      this.logger.error(err);
      throw new Error(`Failed to findChannel with ${_ctoken}.`);
    }

    // check this is the remotelistener of the channel
    if (channel.remotelistenerID != this.remotelistenerEntry._id ) {
      throw new Error(`RemoteListener(${this.remotelistenerEntry._id}) told `+
                      `to doStopChannel ${_args.token}, `+
                      `but channel remotelistenerId is ${channel.remotelistenerID}.`);
    }

    // Turn off remote connection (like webhooks)

    localchanneldata = this.myChannelsLocalData[channel.token];
    this.logger.info('calling doDisableChannel 2');
    await this.doDisableChannel({cref : channel, lref : localchanneldata});
    // this.logger.warn('This next call does not exist. It should basically prep the '+
    //                 'channel for archival purposes.');

    // Post back to Woveon that the channel is not active, and what its icon should be.
    // TODO
//    let result = await this.toWoveon.deactivateChannel(channel.token);
//    // this.logger.verbose(`deactivateChannel result: ${JSON.stringify(result)}`);
    // if ( result.success == false ) {throw new Error(result.error);}

    // remove claim to this channel
    channel.remotelistenerID = null;
    if ( _retainIsActive == false ) {
      channel.isActive = 'inactive'; // by default, mark inactive
      channel.markModified('isActive'); // what a PITA
    }

    // Save channel changes
    await channel.save();

    // remove from self
    delete this.myChannels[_args.token];
    // TODO: myChannels local?

    return WovReturn.retSuccess(true);
  };


  /**
   * Perform any additional authentication with remote service here. Remember to
   * save anything in the channel.
   */
//  async updateCredentials() {};


  /**
   * Get all of own channels, doStopChannels on them, update RemoteListener entry and
   * then stop.
   *
   * Only called when Watcher/shutdown is called, because that slowly drains channels to
   * other RemoteListeners. When it fails the doAliveCheck, just die immediately and kill
   * the pod.
   *
   * @return {string}
   */
  async onShutdown() {

    // shutdown doWake loop
    this.isrunning = false;
    this.logger.aspect('onshutdown', `onShutdown: shutdown remotelistener ${(this.remotelistenerEntry?this.remotelistenerEntry.id:'"not running"')}`);
    // TODO: do doStopChannels in parallel
    let ckeys = Object.keys(this.myChannels);
    this.logger.aspect('onshutdown', ' ... ckeys ', ckeys);
    for (let i=0; i<ckeys.length; i++ ) {
      let token = ckeys[i];
      if ( this.myChannels.hasOwnProperty(token) ) {
        this.logger.aspect('onshutdown', ' ... stop channel (token): ', token);
        await this.doStopChannel(this.myChannels[token]);
        this.logger.aspect('onshutdown', '   ... post stop channel (token): ', token);
      }
    }

    if ( this.remotelistenerEntry != null ) {
      this.logger.aspect('onshutdown', '... delete RL : start ... but may not exist because killed by Overseer');
      await new Promise( function(res, rej) {
        this.logger.aspect('onshutdown', `... delete RL : start internal delete: ${this.remotelistenerEntry.id}`);
        let q = this.RemoteListenerModel.deleteOne({_id : this.remotelistenerEntry.id}, function(err, val) {
          this.logger.aspect('onshutdown', 'deleteOne result: ', err, '2nd : ',  val);
          if (err) rej(err);
          if ( val.ok != 1 ) rej(new Error('Failed database delete call of RemoteListener:'+ val));
          if ( val.n != 1 ) { this.logger.warn(`Failed to delete 1 entry, ${val}, which was probably killed by Overseer. Continuing.`); }
          res(true);
        }.bind(this));

        q.then( (r) => { /* console.log(__filename, 'deleteOne result: ', r); */ res(r); });
        this.logger.aspect('onshutdown', '... delete RL : end internal: ');
      }.bind(this));
    }
    this.logger.aspect('onshutdown', '... delete RL : end');


    // NOTE: since the first thing the loop does is check this variable, it should not screw anything up
    // this.logger.verbose('setting isrunning to false, to wait for loop to stop');
    this.isrunning = false;

    await super.onShutdown();

    return true;
  };


  /**
   * Generator method.
   * @param {object} _configCheck
   * @return {RemoteListenerService} Running/listening object.
   */
  /*
  static async startService( _configCheck) {
    let retval = new RemoteListenerService(_configCheck);
    await retval.init();
    await retval.startup();
    return retval;
  }
  */

};
