

const WEService = require('../service');
const Models    = require('../models');
const ReqWoveon = require('../service/reqWoveon');


/**
 * This handles Res from sipping and listening child classes.
 */
class RemoteResHandler extends WEService {

  /**
   * @param {object} _options -
   */
  constructor(_options = {} ) {
    let options = Object.assign({}, {
      name : 'RemoteDataHandler',
    }, _options);
    super(options);

    // create a specialized model in child class
    this.MessageModel = this.doInitMessageModel();
    if ( this.MessageModel == null ) { this.logger.throwError('You need to implement "doInitMessageModel".'); }

    this.postProcessRes_Messages = null;         // Res post-processing methods... i.e.
    this.IncompleteRes = null;                   // used to connect to incomplete res
    this.toWoveon      = null;                   // requester to Woveon API
  };


    /**
   * initialize this service and routes.
   */
  async onInit() {
    await super.onInit();

    this.IncompleteRes       = Models.incompleteResConnector(this.db);
    await this.MessageModel.connect(this.db);

    // Requester to Woveon API
    this.toWoveon            = new ReqWoveon(this.logger);
  };


  // ---------------------------------------------------------------------
  // Overrides
  // ---------------------------------------------------------------------
  //   Override these methods to implement a remotelistener service for
  //   a specific remote server/service.
  // ---------------------------------------------------------------------

  /**
   * Creates a Resource model for messages of hte remote service. It handles
   * the translation between Woveon and the Remote Service.
   *
   * ex. { return new ResLib.ResModel(
   *                   new ResLib.MessageResProto(),
   *                   new EmailMessagePLProto()); }
   * @return {MessageModel} - a model of messages implemented with Resources.
   */
  doInitMessageModel() { return WEService.WovReturn.retError(null, 'You need to override this method'); }


  /**
   * Connect to the remote service and start being notified of changes.
   * @param {object} _channelref - {ref: channel}, the channel object, only with 'token' so far.
   *                             - Make any changes to the ref.channel object, which will be
   *                               saved after the method exits.
   *                             - Pass ref as _ref.ref so that we can change the contents of channel.
   *                             - args are the params passed directly from WoveonListener
   *                             - if args is null, then channel already has those values (i.e. restarting)
   * @return {array} - Array of subchannels, which are just as what is passed in:
   *   [ {cref : {...}, lref : {...}}, ...]
   *
   * // @return {Long} - The inc of the server. Basically, the server Inc at time of monitoring, to be used for channel.e_mon
   * //                - ex. return new Date().getTime();
   */
  async doConnectChannel(_channelref) {
    try { throw new Error('You need to implement this for this remotelistener.'); }
    catch (err) { console.log(err); console.log('...but continuing, returning cur time...'); }
    // return new Date().getTime();
  };


  /**
   * Disable the remote service notification of changes.
   * @param {object} _channelref - {ref: channel}, the channel object
   */
  async doDisableChannel(_channelref) {
    try { throw new Error('You need to implement this for this remotelistener.'); }
    catch (err) { console.log(err); console.log('...continuing.'); }
  };


  /**
   * Takes the channel data, which could contain multiple messages and other items, and converts them
   * into Res that are fit to send to Woveon.
   *
   * NOTE: if you have to look stuff up, do in post processing
   * NOTE: if you might have missing data, consider it just incomplete data that you will wait around for
   *
   * @param {*} _channeltoken
   * @param {*} _data - array of messages
   * @return {object} - { wovres: array of Resource objects, wovresmods: array of Res that have been modified }
   */
  convertChannelDataToRes(_channeltoken, _data) {
    try { throw new Error('You need to implement this for this remotelistener.'); }
    catch (err) { console.log(err); console.log('...continuing.'); }
    return null;
  };

  // ---------------------------------------------------------------------
  // /Overrides
  // ---------------------------------------------------------------------


  /**
   * When a RemoteService notifies this RemoteListenerService (via webhook,etc.), the
   * data is passed to this method to send to Woveon, and track what data has been pulled
   * from the remote service (in sipBlocks).
   *
   * NOTE: data are in assending inc order
   *
   * @param {object} _channel - the channel
   * @param {array} _msgs - array of Res objects
   * @param {boolean} _doupload - if true, send to server
   * @return {bool} - returns true on successfully handling of all, false means at least one msg failed.
   *                  falase also means no sending.
   */
  async handleNotifiedMessages(_channel, _msgs, _doupload = true) {
    let retval = true;

    // check and send on each message, !!!assuming increasing in time as it goes up!!!
    for (let i=0; i<_msgs.length; i++) {
      if ( ! this.updateSipperWithNotification(_channel, _msgs[i]) )  {
        this.logger.aspect('sippernotification', `handle will return false`);
        retval = false;
      }
    }

    // send data on (set to false for testing)
    // NOTE: if retval is false, all msgs will not be sent... issue? ISSUE? TODO? BUG? ERROR?
    this.logger.aspect('sippernotification', 'Forward to Woveon? :', retval, _doupload);
    if ( retval  && _doupload) {
      await this.handleResForWoveon(_channel.token, _msgs);
      this.logger.aspect('sippernotification', '... handleResForWoveon just finished');
    }

    this.logger.aspect('sippernotification', `handle returning ${retval}`);
    return retval;
  }


  /**
   * Pass all incoming Res to this function to forward to Woveon, save  incomplete,
   * complete/incomplete other Res, etc. This puts the Res into persistent storage,
   * trying to complete it, and when complete, send to Woveon, keeping a trace of
   * the data as needed for future replies and such.
   *
   * This function is used in sipping and listening functions to send Res to Wovoen.
   * On success, creates an entry in the database for the message.
   *
   * Completing/Post-Processing: find the data and apply it to a Res to complete it.
   *
   * for each res,
   *   1. Post-processing: to fetch, compute and correct data (ex. compute a parent); can modify data already sent (PL)
   *   2. Save Modded: Res to Plugin
   *   3. Split: complete and incomplete Res (save incomplete)
   *   4. Completing: try to complete pre-existing incomplete Res in PL with new complete Res
   *   5. Handing Complete: forward Res to Woveon and save
   *   6. Handing Justcompleted: Forward justcompleted and save
   *
   * @param {string} _ctoken - text string identifier of the channel
   * @param {object} _wovres - an array of Res to send. NOTE: actual _wovres saved will differ.
   * @return {integer} - 0 is success, 1 is failed, 2 is failure so turn off channel
   * NOTE: was forwardDataToWoveon
   */
  async handleResForWoveon(_ctoken, _wovres) {
    let retval0 = 1; // modifications to res sent
    let retval1 = 1; // new res
    let retval2 = 1; // just completed res
    let wovresjustcompleted = null;
    let wovrescomplete = null;
    let wovres = null;
    let wovresmods = null;

    this.logger.aspect('handleResForWoveon', '...started handleResForWoveon');
    this.logger.aspect('incres', `...started handleResForWoveon ${_ctoken}`);
    this.logger.aspect('incres', '  wovres: ', _wovres);

    this.logger.aspect('handleResForWoveon', '1: postProcessRes_Messages');
    if ( this.postProcessRes_Messages ) {
      let r = await this.postProcessRes_Messages(_wovres);
      wovres = r.wovres;
      wovresmods = r.wovresmods;

      this.logger.aspect('handleResForWoveon', `2: forwardResToWoveonAndSave ${wovresmods.length}`);
      this.logger.aspect('handleResForWoveon', 'wovresmods: ', wovresmods);
      retval0 = await this.forwardResToWoveonAndSave(wovresmods);
    }
    else { wovres = _wovres; retval0 = 0; }

    this.logger.aspect('handleResForWoveon', `3: handleResForWoveon_PruneIncompleteAndSave ${wovres.length}`);
    wovrescomplete = await this.handleResForWoveon_PruneIncompleteAndSave(wovres); // returns pruned list (on complete)
    // this.logger.aspect('handleResForWoveon', 'wovrescomplete: ', wovrescomplete);
    this.logger.aspect('handleResForWoveon', 'complete: ', wovrescomplete.length);
    // OPTIMIZE: I save to Incomplete here yet some of those will be recreated in the next step. Could just apply without saving incomplete and
    //           save incomplete at end maybe? Then _wovres might actually represent at the end, the Res that were completed.

    this.logger.aspect('handleResForWoveon', `4: handleResForWoveon_CompleteIncompleteWith ${wovrescomplete.length}`);
    // NOTE: this is an ordered list as there are data dependencies since not doing two-phase passes on parents!
    // complete old posts as possible here, with the incoming complete posts
    wovresjustcompleted = await this.handleResForWoveon_CompleteIncompleteWith(wovrescomplete, _ctoken);
    this.logger.aspect('handleResForWoveon', 'just completed: ', wovresjustcompleted.length);
    // this.logger.aspect('handleResForWoveon', 'wovresjustcompleted: ', wovresjustcompleted);

    this.logger.aspect('handleResForWoveon', `5: forwardResToWoveonAndSave ${wovrescomplete.length}`);
    // Forward resources to Woveon API that were originally recieved and save local entry in db.posts
    // TODO add saving to this
    retval1 = await this.forwardResToWoveonAndSave(wovrescomplete);

    this.logger.aspect('handleResForWoveon', `6: handleResForWoveon_ForwardAndSaveJustCompleted ${wovresjustcompleted.length}`);
    // for just completed incomplete posts, forward here (had to wait for parent to forward above) and remove from incomplete list
    retval2 = await this.handleResForWoveon_ForwardAndSaveJustCompleted(wovresjustcompleted);

    let retval = Math.max(retval0, retval1, retval2); // do a max to preserve worst error
    this.logger.aspect('handleResForWoveon', `7: returning ${retval}:  ${retval0} ${retval1} ${retval2}`);

    // _wovres.forEach( (r, i) => r.displayData(`${this.logger.ord(i)}: postforward, in func`) );

    return retval;
  };


  /**
   * Forwards complete Res objects to Woveon, handling errors and saving a minimum amount of data locally.
   *
   * @param {ArrayOfResOrRes} _res - The Res object(s) to send
   */
  async forwardResToWoveonAndSave(_res) {
    let retval = 0;
    let reses = Array.isArray(_res) ? _res : [_res];

    if ( reses.length > 0 ) {

      this.logger.aspect('handleResForWoveon', '  ... bulkUpload ... start', reses.length);
      let result = await this.toWoveon.bulkUpload(reses);
      this.logger.aspect('handleResForWoveon', '  ... bulkUpload result:', result.length);
      // note: the result will contain the ids if needed

      // this.logger.info('bulkUpload result from sipping : ', result.length);
      this.logger.aspect('handleResForWoveon', '... save data back to db here');
      for (let i=0; i<reses.length; i++) {
        let r  = reses[i];  // res
        let rr = result[i]; // result

        // console.log('asdfasdfasdf: ', r);
        // console.log('asdfasdfasdf: ', rr);
        this.logger.info(`Res sent to API: ${r.getXID()} : ${r.getSubject()}`); // r.getreqWovData().data.subject);
        // console.log('asdfasdfasdf2 ', rr.success);
        if ( rr.success == false ) {
          this.logger.error('Failed uploading Res to API.');
          retval |= 1;
          if ( rr.msg == 'Access denied' ) retval |= 2;
        }
        else {
          // save pl data to plugin database (bulk upload above)
          this.logger.aspect('handleResForWoveon', '  ... save updates to database');
          await r.updateMongo();
          this.logger.aspect('handleResForWoveon', '  ... document saved');
        }

      }
    }

    this.logger.aspect('handleResForWoveon', '  ... forwardResToWOveon returning ', retval);
    return retval;
  }


  // ---------------------------------------------------------------------
  // ---------------------------------------------------------------------
  //   ___                                 _      _       ____
  //  |_ _|_ __   ___ ___  _ __ ___  _ __ | | ___| |_ ___|  _ \ ___  ___
  //   | || '_ \ / __/ _ \| '_ ` _ \| '_ \| |/ _ \ __/ _ \ |_) / _ \/ __|
  //   | || | | | (_| (_) | | | | | | |_) | |  __/ ||  __/  _ <  __/\__ \
  //  |___|_| |_|\___\___/|_| |_| |_| .__/|_|\___|\__\___|_| \_\___||___/
  // ---------------------------------------------------------------------
  //  Functions handling Res objects that do not have all the infomation
  //  necessary to send to Woveon. Keep in mongodb db.incompleteres until
  //  complete.
  // ---------------------------------------------------------------------

  /**
   * Pull out incomplete resource objects and saves them as incomplete. Returns the complete ones so they can
   * continue processing in handleResForWoveon.
   *
   * @param {ArrayofRes} _wovres - Array of Res objects, that may have incomplete Res in them.
   * @return {ArrayofRes} - only complete Res objects
   */
  async handleResForWoveon_PruneIncompleteAndSave(_wovres) {
    let retval = [];
    let incomplete = [];

    this.logger.aspect('handleResForWoveon', '...started handleResForWoveon_PruneIncompleteAndSave: numres:', _wovres.length);
    // _wovres.forEach((r)=>r.displayData());

    // look at all res and split by completeness
    for (let i=0; i<_wovres.length; i++) {
      let r = _wovres[i];
      // r.displayData(`all res:  iscomplete(${r.isComplete()})`);
      if ( r.isComplete() == false ) { incomplete.push(r); }
      else { retval.push(r); }
    }
    /* This was destructive. Changing
      for (let i=0; i<_wovres.length; i++) {
        let r = _wovres[i];
        r.displayData(`all res:  iscomplete(${r.isComplete()})`);
        if ( r.isComplete() == false ) {
          let incitem = _wovres.splice(i, 1)[0];
          incomplete.push(incitem);
          i--; // make sure we recheck this index again (since this is destructive to array)
        }
      }
      */

    this.logger.aspect('incres', `  ... found ${retval.length} complete, and ${incomplete.length} Res objects`);

    // save in parallel (i.e. no data dependencies)
    // let savecalls = [];
    for (let i=0; i<incomplete.length; i++) {
      let incres = incomplete[i];
      incres.displayData(`incomplete res:  iscomplete(${incres.isComplete()})`);

      // delete all undefined, since mongodb converts them to null
      let data = {'_wovdata' : incres._wovdata, '_rsdata' : incres._rsdata, 'remoteParentId' : incres.getRemoteParentID()};
      Object.keys(data).forEach((k) => (data[k] === undefined) && delete data[k] );
      Object.keys(data._wovdata).forEach((k) => (data._wovdata[k] === undefined) && delete data._wovdata[k] );
      Object.keys(data._rsdata).forEach((k)  => (data._rsdata[k]  === undefined) && delete data._rsdata[k] );
      // logger.info('saving data to incomplete: ', data);

      // let savepromise =
      let result = await (this.IncompleteRes.findOneAndUpdate(
        {'_wovdata.xid' : incres._wovdata.xid}, data, {upsert : true, new : true}).exec());
      // logger.info('result: ', result);

      // savecalls.push(savepromise);
    }

    // let saveresults =  TODO: should probably do something with the return here to check it saved... try/catch maybe to
    // await Promise.all(savecalls);

    this.logger.aspect('incres', '...ended handleResForWoveon_PruneIncompleteAndSave: ', retval.length);
    return retval;
  }


  /**
   * Repeatedly call handleResForWoveon_incrementalCompleteIncomplete until finished
   * @param {ArrayOfRes} _wovres - The Res that just arrived, so see if we can use them to complete some incompletes.
   * @param {String} _ctoken - the token to index the channel these res are on
   * @return {ArrayOfRes} - all the Res we can complete
   */
  async handleResForWoveon_CompleteIncompleteWith(_wovres, _ctoken) {
    let wovrestocheck = Object.assign([], _wovres); // shallow copy
    let wovresjustcompleted = [];
    this.logger.aspect('incres', '  ...handleResForWoveon_CompleteIncompleteWith: ', _wovres.length);

    while (wovrestocheck.length > 0 ) {
      if ( this.logger.testAspect('incres') ) wovrestocheck.forEach((r) => r.displayData('wovres try to complete res'));
      let rr = await this.handleResForWoveon_incrementalCompleteIncomplete(wovrestocheck, _ctoken);
      this.logger.onAspect('incres', () => rr.forEach( (r) => console.log(r.getIDs(true, true, 'wovres just completed'))));
      wovresjustcompleted = wovresjustcompleted.concat(rr);
      wovrestocheck = rr; // now use these to try and complete others
    }

    // this.logger.onAspect('incres', (l) => wovresjustcompleted.forEach( (r, i) => l.info(`incres: wovres just completed: ${l.ord(i)}`, r.getIDs())));
    return wovresjustcompleted;
  }

  /**
   * Complete posts as possible here
   *
   * @param {ArrayOfRes} _wovres - The Res that just arrived or were completed, so see if we can use them to complete some incompletes.
   * @param {String} _ctoken - the token to index the channel these res are on
   * @return {ArrayOfRes} - Res we have completed with the passed in values (but these may complete others so check again)
   */
  async handleResForWoveon_incrementalCompleteIncomplete(_wovres, _ctoken) {
    let retval = [];
    // let tosave = [];
    this.logger.aspect('incres', '...started handleResForWoveon_CompleteIncompleteWith: ', _wovres.length);

    let lchannel = this.myChannelsLocalData[_ctoken];

    // if res can complete a post, pull that data from db and put in completed res
    for (let i=0; i<_wovres.length; i++) {
      let r = _wovres[i];
      this.logger.info(`see if ${r.getXID()}:${r.getRemoteID()} can complete any incompletes`);
      r.displayData(`res found ${i}`);

      try {
        // NOTE: assumption that incompleteness is due to missing parent id. TODO: make a function on Res that checks if can complete
        let incchildrendata = await this.IncompleteRes.find({'_wovdata.remoteParentId' : r.getRemoteID()});
        logger.info('incchildrendata : ', incchildrendata);

        // build res from the data, and should be complete now
        for (const d of incchildrendata ) {
          // TODO: Optimize by passing in 'r', which is the parent, so it does not need to call DB to find parent

          // build this Res, with data that has already been postprocessed.
          let rs = this.MessageModel.generateRes(d._wovdata, d._rsdata);
          rs.displayData('build res from just completed inc data, iscomplete: '+rs.isComplete());
          logger.info('r: ', r);
          rs.applyPostProcessing( await rs.tryToCompleteWithRes(r) ); // save, so it's values can be used.
          if ( rs.isComplete() ) {
            await rs.updateMongo();
            logger.info(`rs is complete: deleting from incomplete: ${d._wovdata.xid}`);
            retval.push(rs);
            await this.IncompleteRes.deleteMany({_id : d._id}); // TODO: don't need to wait for this
          }
        }
      }
      catch (e) { this.logger.info('continue: ', e); }
    }

    // await Promise.all(tosave);

    this.logger.aspect('incres', '...ended handleResForWoveon_CompleteIncompleteWith: ', retval.length);
    return retval;
  }


  /**
   * For completed posts, forward here in order (had to wait for parent to forward above) and delete
   * the incomplete item in hte incompletepost collection.
   *
   * TODO: This is not the most efficient way of handling this. But, it works for now. This is a slow and steady
   *       process so I am not concernred about speed... yet...
   *
   * @param {ArrayOfRes} _wovresjustcompleted - The list of Res that were recently completed with the addition of
   *    some new Res. The new Res have been sent, so these Res are being sent now and on success, their incomplete
   *    entry in IncompleteRes are then deleted.
   */
  async handleResForWoveon_ForwardAndSaveJustCompleted(_wovresjustcompleted) {
    let retval = 0;

    this.logger.aspect('handleResForWoveon', '... start ForwardAndSaveJustCompleted');

    // NOTE: There are 3 saves: to Mongo.posts database, delete from Mongo.incompleteposts, and send to APIREST. The
    //       Mongo.posts save is above, in handleResForWoveon_incrementalCompleteIncomplete call, when Res is generated.
    //       A good programmer (apparently not me right now, though I claim it's time) would put all three into an atomic
    //       operation with roll-back... FIXME

    // for each res, forward and delete from incomplete if successful
    for (let i=0; i<_wovresjustcompleted.length; i++) {

      let r  = _wovresjustcompleted[i];
      this.logger.aspect('handleResForWoveon', '  r is ', r.getXID());

      // forward and save again (inside forwardResToWoveonAndSave)
      let forwardresult = await this.forwardResToWoveonAndSave(r);
      this.logger.aspect('handleResForWoveon', '  r forward is  ', forwardresult);
      retval |= forwardresult;

      // if success, delete incomplete res entry : NOTE: removed this since IncRes are removed in CompleteIncomplete
      /*
      if ( forwardresult == 0 ) {
        let delresult     = null;
        this.logger.aspect('handleResForWoveon', '  delete xid ', r.getXID());
        delresult = await this.IncompleteRes.deleteOne({'_wovdata.xid' : r.getXID()});
        this.logger.aspect('handleResForWoveon', '  r delresult is  ', delresult);
        this.logger.info('delresult is "', delresult, '" ... I have no idea what remove returns... so handle how?');
        if ( ! (delresult.ok == 1 && delresult.n == 1) ) retval |= 1; // if not ok and deleted 1 (n==1), then error
      }
      */
    }

    this.logger.aspect('handleResForWoveon', '... end ForwardAndSaveJustCompleted ', retval);

    return retval;
  }


  // ---------------------------------------------------------------------
  // /IncompleteRes
  // ---------------------------------------------------------------------

  /*
   * Postprocessing: Additionally, do operations to fetch data/pre compute data, as per
   *    the specifc Remote System needs. Generally, the RS protocol is lacking some data
   *    in what it sends that prevents knowing it's parent. Usually, an RS that is not
   *    hierarchical but time based, would need to look up 'last in that converation'
   *    and/or set parent from the passed in messages
   *
   * !!!NOTE: Also request any comments or likes that are needed to full fulfull
   * the post.
   */

  /**
   * For most Messages, we just need to compute: xidTop, xidParent, remoteId, remoteParentId
   * NOTE: lack of a parent will leave these as incomplete, which is fine
   * @param {arrayOfRes} _wovres - array of newly created Res
   * @return {object} - {wovres : array of new Res, wovresmods : existing Res that are modified}
   */
  async postProcessRes_Messages_Hierarchy(_wovres, _data = {}) {

    this.logger.aspect('handleResForWoveon', '...postProcessRes_Messages_Hierarchy - start');

    for (let i=0; i<_wovres.length; i++) {
      let r = _wovres[i];

      // get will set wovdata
      r.getResInc();       // compute
      r.getRemoteID();     // compute

      // looks up parent and such
      r.applyPostProcessing( await r.tryToCompleteWithRes(null) ); // save, so it's values can be used.

      /*
      // if we have a parent, fetch it. or otherwise try to set xidparent or xidtop
      let rpid = r.getRemoteParentID();
      // logger.info('rpid : ', rpid);
      if ( rpid != null ) {
        let parent = null;

        if ( _data.parent !== undefined ) parent = _data.parent;
        else parent = await r._resmodel.models.findOne({remoteId : rpid}).lean();

        // logger.info('parent: ', parent);
        if ( parent != null ) {
          r._wovdata.xidParent = parent.xid;
          r._wovdata.xidTop    = parent.xidTop;
        }
        // logger.info('r is complete?: ', r.isComplete());
      }
      else { r.getXIDParent(); r.getXIDTop(); }
      */
    }

    this.logger.aspect('handleResForWoveon', '...postProcessRes_Messages_Hierarchy - end : ');
    return {wovres : _wovres, wovresmods : [] };
  }


  /**
   * This is run after convertChannelDataToRes by the plugins that need it. It fills in xidTop and xidParent of
   * messages that don't have that data, based upon:
   *
   * ASSUMPTION: This is for non-heirarchical remote server messages, for reordering xidTop and xidParent.
   * ASSUMPTION: This uses a stub message for xidTop. It also sets parents when Inc values are screwy.
   * ASSUMPTION: It assumes no incomplete, since we never know if we got all messages in a
   *             conversation as we fetch them so instead we send modifications in place of incomplete
   *
   * @param {arrayOfRes} _wovres - array of newly created Res
   * @return {object} - {wovres : array of new Res, wovresmods : existing Res that are modified}
   */
  async postProcessRes_Messages_NonHierarchy(_wovres) {
    let retval = {
      wovres     : _wovres.slice(0),  // all the new Res (sorted)
      wovresmods : [],                // all the old Res that were modified (unordered)
    };
    let spanningres = null;           // Res that exist in DB already (sorted)
    let allres = _wovres.slice(0);    // All the Res, from spanning and _wovres and lmsgs/stub. (sorted)

    let smsg = null;                  // Stub Msg. track this. need to send stub to WOVEON, but not existing wovres[0]
    let lmsg = null;                  // left bound Res
    let xidtop = null;

    // NOTE: because incoming messages could be same time, and we use random bits and counter for Inc, then the 'right msg',
    //       already in system, may have Inc before. So, make sure to order by Inc.
    // TODO: to future me. reorder Inc if you can tell order by arrival. but, that has issues too. expecting odds of this to be very, very, very low

    this.logger.aspect('transformres', 'step1: (left msg)  find earliest msg in conversation that is before first wovres Res.');
    lmsg = await this.getLeftMostMsg(_wovres[0]);
    if ( lmsg != null && lmsg.getResInc().greaterThan(_wovres[0].getResInc()) ) lmsg = null;
    this.logger.aspect('transformres', '   lmsg: ', (lmsg != null? JSON.stringify(lmsg._wovdata):null));

    this.logger.aspect('transformres', 'step1: if no left msg (i.e. no messages in conversation yet), create stub for left msg. Prepend either to wovres, sorted.');
    if ( lmsg ) {
      xidtop = lmsg.getXIDTop();                   // get xidTop
      retval.wovresmods.push(lmsg);                // lmsg will be modified (parent)
      allres = lmsg.insertIntoSortedArray(allres); // lmsg anchors (most likely) allres
    }
    else {
      smsg = this.genStubRes(_wovres[0]);          // create stub msg
      xidtop = smsg.getXIDTop();                   // get xidTop
      retval.wovres.unshift(smsg);                 // smgs is new
      allres.unshift(smsg);                        // smsg by definition is left of all
    }
    this.logger.aspect('transformres', '   smsg: ', (smsg != null? JSON.stringify(smsg._wovdata):null));
    for (let i=0; i<retval.wovres.length; i++) { logger.info(`  - ${logger.ord(i)} res: ${JSON.stringify(retval.wovres[i]._wovdata)}`); }

    this.logger.aspect('transformres', 'get right bound inc ( less than message to right of last message in wovres or if rmsg is null, use date from wovres[wovres.length-1] date)  ');
    let rboundinc = null;
    let rmsg = await this.getRightMsg(_wovres[_wovres.length-1]); // message to right of last wovres message


    if ( rmsg != null ) {
      rboundinc = rmsg.getResInc();
      rmsg.displayIDs('rmsg for rbound');
    }
    else {
      rboundinc = _wovres[_wovres.length-1].getResInc();
    }
    this.logger.aspect('transformres', 'rmsg for rbound: ', rboundinc);

    this.logger.aspect('transformres', 'allres now');
    allres.forEach( (r) => r.displayIDs() );

    this.logger.aspect('transformres', 'spanning block: get block of messages that span the block of wovres so we can update all parents');
    // here... find from wovres[0] < date < rmsg and insert into this list so we can adjust parents of all
    // NOTE: this is to remove assumptions that messages are arriving in proper order
    spanningres = await this.getResIncSpan(_wovres[0], allres[0].getResInc(), rboundinc);
    this.logger.aspect('transformres', 'spanning ', spanningres.length);
    spanningres.forEach( (r) => r.displayIDs() );
    spanningres.forEach( (r) => retval.wovresmods.push(r) );

    this.logger.aspect('transformres', 'step1: put spanningres into allres, sorted, and retval.wovresmods, unsorted ');
    for (let i=0; i<spanningres.length; i++) { allres = spanningres[i].insertIntoSortedArray(allres); }
    allres.forEach( (r) => r.displayIDs() );

    this.logger.aspect('transformres', 'step1: if rmsg, this needs to point at last wovres so put into allres at end (and add to mods list)');
    if ( rmsg != null ) { allres = rmsg.insertIntoSortedArray(allres); retval.wovresmods.push(rmsg); }

    this.logger.aspect('transformres', 'allres now');
    allres.forEach( (r) => r.displayIDs() );

    this.logger.aspect('transformres', 'r.wovresmods now');
    retval.wovresmods.forEach( (r) => r.displayIDs() );

    this.logger.aspect('transformres', 'step2: set xidTop of retval.wovres (i.e. all new Res, set to stub or lmsg value from above) : ', xidtop);
    allres.forEach( (r) => r._wovdata.xidTop = xidtop );

    this.logger.aspect('transformres', 'step2: if lmsg (i.e no stub), set xidParent of wovres[0] to leftmsg.xidParent (since no longer left).');
    this.logger.aspect('transformres', '       Handles case of lmsg not being left most message.');
    // NOTE/TODO: if lmsg is allres[0], no need to set its parent AND no reason to put into retval.wovresmods
    if ( lmsg ) retval.wovres[0]._wovdata.xidParent = lmsg.getXIDParent();

    this.logger.aspect('transformres', 'step2: reorder parents: set xidParent of wovres[i] to wovres[i-1] (i.e. stub, or existing)');
    for (let i=1; i< allres.length; i++) {
      allres[i]._wovdata.xidParent = allres[i-1]._wovdata.xid;
      allres[i]._wovdata.remoteParentId = allres[i-1]._wovdata.remoteId;
    }


    /*
     * removed since this deals with mod vs new Res
      this.logger.aspect('transformres', 'TODO step3: if rmsg, pop from wovres and add to wovresmods');
      if ( rmsg ) {
        let p = rmsg.findPosInArray(retval.wovres);
        retval.wovres.splice(p, 1);
        retval.wovresmods.push(rmsg);
      }

      this.logger.aspect('transformres', 'step3: pop left msg from wovres. if lmsg was not wovres[0], add to wovresmods');
      if ( lmsg ) {
        let p = lmsg.findPosInArray(retval.wovres);
        retval.wovres.splice(p, 1);
        if ( p != 0 ) retval.wovresmods.push(lmsg); // i.e. modified xidParent
      }
      */

    this.logger.h3('transformresend');
    this.logger.aspect('transformresend', 'allres'); allres.forEach( (r) => r.displayIDs() );
    this.logger.aspect('transformresend', 'lmsg'); if ( lmsg ) lmsg.displayIDs();
    this.logger.aspect('transformresend', 'smsg'); if ( smsg ) smsg.displayIDs();
    this.logger.aspect('transformresend', 'all new Res'); retval.wovres.forEach( (r) => r.displayIDs() );
    this.logger.aspect('transformresend', 'all modified Res'); retval.wovresmods.forEach( (r) => r.displayIDs() );

    return retval;
  }

};

module.exports = RemoteResHandler;

