
const WEService     = require('../service'); // NOTE: WEService
const _ChannelModel = require('../models/channelConnector');

let C = null;


module.exports = class WoveonListenerService extends WEService {


  /**
   * Constructor.
   * @param {object} _options -
   *   : staticdir - if this serves static content, set the directory to use (ex. 'static'
   */
  constructor(_options = {}) {
    C             = require('woveon-service').Config;
    let options = Object.assign({}, {
      staticdir             : null,
      baseroute             : `/${C.get('WOV_we_ver')}`,
      channelValidateLaunch : undefined,
      name                  : 'WoveonListener',
      port                  : C.get('WOV_we_wl_port'),
      logger                : C.logger,
    }, _options);
    super(options);
    if ( this.logger.testAspect('config.wl') ) C.displayMe();
    this.toWoveon = null; // set in onInit

    // Create the Resource model for messages of the remote service
    this.MessageModel = this.doInitMessageModel();
    if ( this.MessageModel == null ) {
      this.logger.throwError('You need to implement "doInitMessageModel".');
    }
  }


  // ---------------------------------------------------------------------
  // Overrides
  // ---------------------------------------------------------------------
  //   Override these methods to implement a remotelistener service for
  //   a specific remote server/service.
  // ---------------------------------------------------------------------


  /**
   * Override this to redirect to where the plugin gathers and validates access to the RemoteService.
   * Should contact the remote service and authenticate with it. Then eventually call /channel_launch.
   *   - ex. Facebook, contacts it
   *   - ex. Email, use forms to collect email account data then verify it works on the IMAP server
   *   NOTE: was remoteAuth
   *
   * @param {object} _args - { t:, } where t is the channel's token.
   * @param {object} _res - http response object (in case you need it) (_res.req.originalUrl is useful)
   * @return {WovReturn} - redirect
   */
  async onChannelStart(_args, _res) {
    let msg = `This method should be overriden if remote server needs some action.`;
    this.logger.error(msg);
    return WEService.WovReturn.retFail(null, 400, msg);
  };


  /**
   * Override this handler for checking data (i.e. user inputed form data) before creating the channel. On
   * success, will call channel_launch.
   * @param {object} _channelModel - A Mongoose model of the channel.
   * @param {object} _args - Params to check against RemoteService.
   * @return {WovReturn} - null on success, error if data fails
   */
  async onValidateLaunch(_channelModel, _args) { return WEService.WovReturn.retError(_args, 'This method needs to be overridden.'); }


  /**
   * During the /channel_launch, this function is called for the plugin to take the data
   * entered from the user or passed from the RemoteService to save for the channel. This
   * data is necessary to communicate with the RemoteService, like tokens, passwords, account
   * names, etc.
   * @param {object} _channeldata - the raw values passed to /channel_launch
   * @return {WovReturn} - with data being the pluginData to save to the mongo database for the channel.
   */
  packPluginData(_channeldata) {
    let msg = `This method should be overriden to create the pluginData object. ex. retSuccess({access_key: ''})`;
    this.logger.error(msg);
    return WEService.WovReturn.retError(null, msg);
  };


  /**
   * Creates a Resource model for messages of hte remote service. It handles 
   * the translation between Woveon and the Remote Service.
   *
   * ex. { return new ResLib.ResModel( 
   *                   new ResLib.MessageResProto(), 
   *                   new EmailMessagePLProto()); }
   */
  doInitMessageModel() {
    return WEService.WovReturn.retError(null, 'You need to override this method');
  }

  /**
   * Override these event callbacks as needed by your plugin...
   *
   * @param {Res} _res - A resource object, created from data passed from the Api.
   * @return {WovReturn} - Null on success, or WovReturn in error/fail state.
   */
  /*
  async onCreatePost(Res);    // {token:, message:,}
  async onUpdatePost(Res);    // {token:, postId:, message:,}
  async onDeletePost(Res);    // {token:, postId:,}
//  async onCreateComment(Res); // {token:, postId:, message:,}
//  async onUpdateComment(Res); // {token:, commentId:, message:,}
//  async onDeleteComment(Res); // {token:, commentId:}
  async onLike(Res);          // {postId:, commentId:,}
  async onUnlike(Res);        // {postId:, commentId:,}
  */

  // ---------------------------------------------------------------------
  // /Overrides
  // ---------------------------------------------------------------------


  /**
   * Handle messages from the API, and route them to the remote service.
   * @param {object} _args - { type:, token:, etc...}
   *   NOTE: token is always the channel token
   * @return {object} call-specific data is returned
   *    createPost   { postId:, authorUrl:, }
   */
  async onAPIEvent(_args) {
    this.logger.info('onApiEvent, Passing in args: ', JSON.stringify(_args, null, '  '));

    let retval = WEService.WovReturn.checkAttributes(_args, ['token', 'type'], null, {checkStrict : false});
    this.logger.info('onAPIEvent : ', JSON.stringify(retval, null, '  '));
    let channel = null;

    if ( retval == null ) {
      this.logger.aspect('onApiEvent', 'Get the channel');
      channel = await this.ChannelModel.findOne({token : _args.token});
      if (channel == null) {
        retval = WEService.WovReturn.retError(`API asked for 'createComment' to unknown channel of token ${_args.token}.`);
        this.logger.error(retval);
      }
    }

    if ( retval == null ) {
      let result = null;

      // generate the res, but no reason to save since RL handles that (simplifies WL/RL collaboration)
      let res = await this.MessageModel.generateResWL(_args, null);
      // await res.updateMongo();

      if ( _args.type == 'updatePost' && this.onUpdatePost ) {
        result = await this.onUpdatePost(_args);
      }
      else if ( _args.type == 'deletePost' && this.onDeletePost ) {
        result = await this.onDeletePost(_args);
      }
      else if (_args.type == 'createPost' && this.onCreatePost) {

        let parentmodel = null;
        if ( this.logger.testAspect('onCreatePost') ) res.displayData('createPost with this res');
        if ( res.getXIDParent() != null ) {
          this.logger.aspect('onCreatePost', 'get parent message : ', res.getXIDParent());
          parentmodel = (await this.MessageModel.models.findOne({xid : res.getXIDParent()}).lean());
          this.logger.aspect('onCreatePost', 'parent: ', parentmodel);
          // If parentmodel is null, this is a new message.
          /*
          if ( parentmodel == null ) {
            res.displayData('No parent to this message, which is weird since this is a reply hopefully');
            retval = WovReturn.retError(`Could not find parent message: ${res.getXIDParent()} for msg ${res.getXID()}.`);
          }
          */
        }

        if ( retval == null ) {
          result = await this.onCreatePost(res, channel.pluginData, parentmodel);
          this.logger.aspect('onCreatePost', '  ...result: ', result);
        }
      }
      else if ( _args.type == 'updateComment' && this.onUpdateComment ) {
        result = await this.onUpdateComment(_args);
      }
      else if ( _args.type == 'deleteComment' && this.onDeleteComment) {
        result = await this.onDeleteComment(_args);
      }
      // TODO: HERE PUT LIKE handlers
      else {
        let e = new Error('Unhandled API Event:', _args);
        retval = WEService.WovReturn.retError(e, 'Unhandled API event ', _args.type);
      }

      if ( retval == null ) retval = WEService.WovReturn.retSuccess(result);
    }

    this.logger.info('returning: ', JSON.stringify(retval, null, '  '));

    return retval;
  };


  /**
   * initialize this service and routes.
   */
  async init() {
    await super.init();


    // allow redirect to app
    this.logger.info('CORS setting access control to all in WoveonService... so removing this');
    /*
    this.listener.app.use(function(req, res, next) {
      res.header("Access-Control-Allow-Origin", "*");
      res.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept");
      next();
    });
    */

    this.toWoveon     = new WEService.ReqWoveon(this.logger);
    this.ChannelModel = _ChannelModel(this.db);

    // connect the message model
    this.MessageModel.connect(this.db);

    // /app GET - redirect to app
    // ---------------------------------------------------------------------
    this.listener.onGet('/app', (async function() {
      return WEService.WovReturn.retRedirect(`${C.get('WOV_www_app_fullurl')}`);
    }), __filename);


    // /channel_auth GET
    // ---------------------------------------------------------------------
    this.listener.onGet('/channel_auth', this.channelAuth, __filename, {
      summary   : 'API is calling because user wants a new channel.',
      desc      : 'Creates a channel that is starting and has a token identifier, that is returned and used to start the channel.',
      params    : [],
      responses : {
        '200' : {
          desc   : 'Success',
          params : [
            {name : 'url', desc : 'The full URL to call', Type : 'string'},
            {name : 'token', desc : 'A randomized token to refer to this channel.', Type : 'string'},
            {name : 'isActive', desc : 'Always false.', Type : 'boolean'},
          ],
        },
      },
    });


    // /channel_deauth DELETE
    // ---------------------------------------------------------------------
    this.listener.onDelete('/channel_deauth', this.channelDeauth, __filename, {
      summary : 'Pass to msRemoteListener to handle deletion and deregistration from remote service.'});


    // /channel_start POST
    // ---------------------------------------------------------------------
    let cs_doc = {
      summary : 'Process to start a newly created channel',
      desc    : 'User call here to begin authenticating and setting up with the remote service. This will first redirect to:\n'+
                '  - a RemoteService to auth and gather data to start the channel\n'+
                '  - a static webpage with a form to gather data from the user\n'+
                '  - for pltest, the app, assuming all is successful.\n'+
                'Returns to /channel_launch, which then leads to the app on success or an error page on failure.\n',
      params    : [],
      responses : {'302' : {desc : 'redirection'}},
    };
    this.listener.onGet('/channel_start',  this.channelStart, __filename, cs_doc);
    this.listener.onPost('/channel_start', this.channelStart, __filename, cs_doc);


    // /channel_launch and /channel_validatelaunch
    // ---------------------------------------------------------------------
    // Enable the correct launch URL for this channel based upon validation needs.
    // SECURITY NOTE: channelValidateLaunch calls channelLaunch on success and by not enabling
    //                channelLaunch, can't enable with bad params.

    if (this._options.channelLaunch == true) {
      this.listener.onGet('/channel_launch', this.channelLaunchHTML, __filename);
      if ( C.get('WOV_STAGE') == C.get('WOV_ME') || C.get('WOV_STAGE') == 'dev' ) {
        this.listener.onGet('/channel_launch_', this.channelLaunchJSON, __filename);
      }
      else { this.logger.info('...not activating internal routes: /channel_launch_'); }

    } else { this.logger.info('skipping /channel_launch'); }

    if (this._options.channelValidateLaunch == true) {
      this.listener.onPost('/channel_validatelaunch', this.channelValidateLaunchJSON, __filename);
      this.listener.onGet('/channel_validatelaunch', this.channelValidateLaunchJSON, __filename);
    }
    else { this.logger.info('skipping /channel_validatelaunch'); }

    // else { throw Error('channelValidateLaunch is required and needs to be true or false'); }

    // /channel_validatelaunch POST
    // ---------------------------------------------------------------------
    this.listener.onDoc('/channel_validatelaunch', {
      summary : 'Test values and return WovReturn error on error, or on success, it calls and returns the '+
                'values from this.channelLaunch. See /channel_launch for params.',
      responses : {
        '400' : {desc : 'Returns when validation error.'},
        '302' : {desc : 'Returns the redirect from this.channelValidateLaunch on success'},
      },
    });

    // /channel_launch POST
    // ---------------------------------------------------------------------
    // TODO: how about other errors? need to redirect to a standard page (i.e. Facebook failed or plugin failure, try again)
    this.listener.onDoc('/channel_launch', {
      summary : 'Tell the plugin to launch the channel with the given data.',
      desc    : 'This should be called with all the appropriate data for the channel to run and talk to the RemoteService.'+
                'Calls the onValidateLaunch if the plugin uses it.',
      params : [
        {name : 'token', desc : 'The identifying token for this channel.', required : true},
        {
          name     : 'pageImage',
          desc     : 'The on-page image for this channel (i.e. user\'s icon). When null, one will be generated.',
          required : false,
        },
      ],
      responses : {
        '302' : {desc : 'All success. Redirect back to app. RemoteListener will grab and run the channel shortly.'},
      },
    });


    // /hook POST
    // ---------------------------------------------------------------------
    this.listener.onPost('/channel_hook', this.onAPIEvent, __filename, {
      summary : 'Listen for messages from the API and forward to remote service.',
    });

  };


  // ---------------------------------------------------------------------
  //   _   _                 _ _
  //  | | | | __ _ _ __   __| | | ___ _ __ ___
  //  | |_| |/ _` | '_ \ / _` | |/ _ \ '__/ __|
  //  |  _  | (_| | | | | (_| | |  __/ |  \__ \    Listeners forward to
  //  |_| |_|\__,_|_| |_|\__,_|_|\___|_|  |___/     the methods below...
  // ---------------------------------------------------------------------

  /**
   * Create an entry for a new Channel. API is calling.
   * @param {object} _args - empty
   * @return {object} - url for user to contact plugin at, token of this new channel
   */
  async channelAuth(_args) {
    this.logger.aspect('wl.channel.auth', 'trying to Auth channel ', _args);
    let retval = null;

    // create the channel
    const channel = await new this.ChannelModel({token : WEService.generateToken(), pluginData : {}}).save();
    if ( channel == null ) { retval = WEService.WovReturn.retError(null, 'Failed to auth a channel'); }

    if ( retval == null ) {

      // create the base URL... if localhost, add port, otherwise no port
      // let baseurl = C.get('WOV_we_wl_url');
      let baseurl = C.get('WOV_we_wlexternal_url');
      if ( baseurl.includes('localhost') ) baseurl += `:${C.get('WOV_we_wl_port')}`;

      // check for protocol
      if ( ! baseurl.startsWith('http') ) { this.logger.throwError('WOV_we_wlexternal_url must start with protocol (i.e. http(s)).'); }

      // create the return object
      retval = WEService.WovReturn.retSuccess({
        url      : `${baseurl}${this._options.baseroute}/channel_start?t=${channel.token}`,
        token    : channel.token,
        isActive : false, // TODO: why? how is this used? always false...
      });
      this.logger.info('channelAuth returning start url (through API, and on to browser): ', retval.data.url);
    }

    return retval;
  };


  /**
   * Forward to the msRemoteListener. It will delete a token and deregister with external service.
   * TODO: if RL keeps data in memory, this won't be reread
   *
   * @param {*} _args -
   * @return {bool} success
   */
  async channelDeauth(_args) {
    this.logger.aspect('wl.channel.deauth', 'trying to Deauth channel ', _args);
    let retval = WEService.WovReturn.checkAttributes(_args, ['token']);
    let channel = null;

    if ( retval == null ) {
      this.logger.aspect('wl.channel.deauth', 'find channel');
      channel = await this.ChannelModel.findOne({token : channelToken});
      this.logger.aspect('wl.channel.deauth', 'channel to deauth: ', channel);
      if ( channel == null ) { retval = WEService.WovReturn.retError(_args, 'Trying to deAuth a Channel that does not exist!'); }
    }

    if ( retval == null ) {
      this.logger.aspect('wl.channel.deauth', 'save channel');
      channel.isActive = 'stopped';
      channel.markModified('isActive');
      await channel.save().catch( (err) => { retval = WEService.WovReturn.retError(_args, 'Error saving'); });
    }

    if ( retval == null ) { retval = WEService.WovReturn.retSuccess(true); }

    return retval;
  }


  /**
   * Checks params and then calls user-defined onChannelStart to generate redirect WovReturn.
   * @param {object} _args - has 't' which is the channel token
   * @return {WovReturn} - a redirect to RS or a static html form page
   */
  async channelStart(_args, _res) {
    let retval  = WEService.WovReturn.checkAttributes(_args, ['t']);
    let channel = null;

    if ( retval == null ) {
      channel = await this.ChannelModel.findOne({token : _args.t});
      // this.logger.info('channel: ', channel);
    }

    if ( retval == null ) {
      if ( channel ) { retval = await this.onChannelStart(_args, _res); }
      else { retval = WEService.WovReturn.retRedirect(`/static/channel_start_error.html`, false); }
    }

    this.logger.aspect('ChannelStart', 'channelStart returning: ', retval);

    return retval;
  }


  /**
   * For plugin validated launch values, call this instead of channelLaunch. Validates and if error, returns error.
   * Calls this.channelLaunch on success, returning a redirect.
   *
   * @param {object} _args -
   * @return {WovReturn} - an error or on success, a url from this.channelLaunch on success
   */
  /*
  async channelValidateLaunchHTML(_args, _res) {
    // this.logger.info('res cvlhtml: ', _res);
    let retval = await this.channelValidateLaunch_internal(_args);
    this.logger.info('retval of validate launch internal: ', retval);
    if ( retval.retval == null ) {
      this.logger.aspect('wl.channel.launch', 'launch channel');
      retval.retval = await this.channelLaunchHTML(_args, null, retval.channel);
    }
    else {
      this.logger.aspect('wl.channel.launch', '... retval error is ', JSON.stringify(retval.retval, null, '  '));
      retval.retval = WEService.WovReturn.retRedirect(`/static/channel_validate_error.html`, false);
    }
    return retval.retval;
  }
  */


  /**
   * @param {object} _args -
   * TODO fix me
   */
  async channelValidateLaunchJSON(_args) {
    let retval = null;
    try {
      retval = await this.channelValidateLaunch_internal(_args);
      if ( retval.retval == null ) {
        this.logger.aspect('wl.channel.launch', 'launch channel');
        retval.retval = await this.channelLaunchJSON(_args, null, retval.channel);
      }
      else {
        this.logger.aspect('wl.channel.launch', '... retval error is ', JSON.stringify(retval.retval, null, '  '));
        // retval.retval = WEService.WovReturn.retError(retval.retval);
        retval.retavl = retval.retval;
      }

      this.logger.aspect('wl.channel.launch', 'returning :', retval.retval);
      this.logger.info('returning :', retval.retval);
    }
    catch (err) {
      this.logger.info('channelValidateLaunchJSON caught error');
      this.logger.info(err);
      this.logger.error(err);
    }
    return retval.retval;
  }


  /**
   * Call by two funcs above.
   * @param {object} _args - {channel: ..., retval: WovReturn }
   */
  async channelValidateLaunch_internal(_args) {
    let retval = {channel : null, retval : null};
    this.logger.aspect('wl.channel.launch', 'channelValidateLaunch: ', _args);
    retval.retval = WEService.WovReturn.checkAttributes(_args, {'token' : true, 'pageImage' : false}, null, {checkStrict : false});

    if ( retval.retval == null ) {
      this.logger.aspect('wl.channel.launch', 'find channel: ', _args.token);
      retval.channel = await this.ChannelModel.findOne({token : _args.token})
        .catch( (err) => { retval.retval = WEService.WovReturn.retError(_args.token, 'could not connect to channel models'); });
      this.logger.info('found channel is: ', retval.channel);
    }
    if ( retval.retval == null ) {
      this.logger.aspect('wl.channel.launch', 'check if channel found');
      if ( retval.channel == null ) { retval.retval = WEService.WovReturn.retError(_args.token, 'could not find channel in plugin database.'); }
    }

    if ( retval.retval == null ) {
      this.logger.aspect('wl.channel.launch', 'validating params');
      retval.retval = await this.onValidateLaunch(retval.channel, _args);
      this.logger.info('validating params result: ', retval);
    }

    this.logger.info('channelValidateLaunch returning : ', retval.retval);

    return retval;
  }


  /**
   * Manages the end of the /channel_start call. Here, we save the channel data to the plugin, activate
   * the channel with Woveon, start the channel in the RemoteListener and redirect back to the UI.
   *
   * NOTE: this used to return {object} { channel:, url: config.app.url }
   *
   * @param {object} _args - {token : channel token, pageImage, ...} - where '...' is specific to the individual plugin
   * @param {response} _res - http Response object (just ignore it)
   * @param {object} _channel - null by default, but channelValidateLaunch passes in a Mongoose model
   * @return {WovReturn} - a redirect or error ...
   *      TODO: it can return a WovReturn error, but RemoteServices would have no way to handle it. need those errors
   *            to be redirect to a page that manages that error
   */
  async channelLaunchHTML(_args, _res, _channel = null) {
    let retval = await this.channelLaunch_internal(_args, _res, _channel);
    if ( retval == null ) {
      this.logger.aspect('wl.channel.launch', '... success ', C.get('WOV_www_app_fullurl'));
      retval = WEService.WovReturn.retRedirect(C.get('WOV_www_app_fullurl'));
    }
    this.logger.info('channelLaunchHTML returning: ', retval);
    return retval;
  };


  /**
   * @param {object} _args - {token : channel token, pageImage, ...} - where '...' is specific to the individual plugin
   * @param {response} _res - http Response object (just ignore it)
   * @param {object} _channel - null by default, but channelValidateLaunch passes in a Mongoose model
   */
  async channelLaunchJSON(_args, _res, _channel = null) {
    let retval = await this.channelLaunch_internal(_args, _res, _channel);
    if ( retval == null ) {
      this.logger.aspect('wl.channel.launch', '... success ', C.get('WOV_www_app_fullurl'));
      retval = WEService.WovReturn.retSuccess({url : C.get('WOV_www_app_fullurl')});
    }
    this.logger.info('channelLaunchJSON returning: ', retval);
    return retval;
  };


  /**
   * call by two funcs above.
   * @param {object} _args - {token : channel token, pageImage, ...} - where '...' is specific to the individual plugin
   * @param {response} _res - http Response object (just ignore it)
   * @param {object} _channel - null by default, but channelValidateLaunch passes in a Mongoose model
   */
  async channelLaunch_internal(_args, _res, _channel = null) {
    let channel = _channel;

    this.logger.aspect('wl.channel.launch', 'channelLaunch: ', _args, _channel);
    let retval = WEService.WovReturn.checkAttributes(_args, {'token' : true, 'pageImage' : false}, null, {checkStrict : false});
      // --- 'token' has to exist, but the rest is plugin dependent

    if ( retval == null && channel == null) {
      this.logger.aspect('wl.channel.launch', 'find channel');
      channel = await this.ChannelModel.findOne({token : _data.token})
        .catch( (err) => { retval = WEService.WovReturn.retError(_args.token, 'could not connect to channel models'); });
    }

    if ( retval == null ) {
      this.logger.aspect('wl.channel.launch', 'check if channel found');
      if ( channel == null ) { retval = WEService.WovReturn.retError(_args.token, 'could not find channel'); }
    }

    if ( retval == null ) {
      this.logger.aspect('wl.channel.launch', 'pack pluginData');
      try {
        let result = this.packPluginData(_args);
        this.logger.aspect('wl.channel.launch', 'pluginData result: ', result);
        if ( result == null ) retval = WEService.WovReturn.retError(_args, 'packPluginData returned null');
        else if ( result.success == false ) {
          this.logger.aspect('wl.channel.launch', 'packPluginData failed ', result);
          retval = result;
        }
        else {
          channel.pluginData = result.data;
          channel.isActive   = 'inactive'; // from 'starting'... RL will grab and start in time
          channel.markModified('pluginData');
          channel.markModified('isActive');
        }
      }
      catch (err) {
        this.logger.warn(err);
        retval = WEService.WovReturn.retError(_args, 'packPluginData failed with these args');
      }
    }

    if ( retval == null ) {
      this.logger.aspect('wl.channel.launch', 'save channel pluginData, and set state to ', channel.pluginData);
      await channel.save()
        .catch(function(err) {
          if (err) {
            this.logger.error(' ... failed saving channel\'s pluginData to database ', _args.token);
            retval = WEService.WovReturn.retError(_args.token, 'failed saving channel\s pluginData to database');
          }
        }.bind(this));
    }

    if ( retval == null ) {
      this.logger.aspect('wl.channel.launch', 'activate channel on API');

      try {
        // Post back to Woveon that the channel is active, and what its icon should be.
        let result = await this.toWoveon.activateChannel(_args.token, _args.pageImage);
        this.logger.info(`toWoveon.activateChannel returned : ${JSON.stringify(result)}`  );
        if ( result.success == false ) {
          let msg = `toWoveon.activateChannel ${_channel.token} failed.`;
          this.logger.error(msg);
          retval = WEService.WovReturn.retError(result, msg);
        }
      }
      catch (err) {
        let msg = `failed call to activateChannel: ${_channel.token}.`;
        this.logger.error(msg+'\n'+err);
        retval = WEService.WovReturn.retError(err, msg);
      }
    }

    this.logger.info('channelLaunch returning: ', retval);
    console.trace();
    return retval;
  };


  /**
   * create the base URL... if localhost, add port, otherwise no port
   *   ?? encodeURIComponent ???
   * @param {string} _ep - trailing endpoint
   * @return {string} - url ex. _ep == '/hi' returns http://localhost:50105/v1/hi or https://app.woveon.com/v1/hi
   */
  genWLURL(_ep) {
    let retval = C.get('WOV_we_wl_url');
    if ( retval.includes('localhost') ) retval += `:${C.get('WOV_we_wl_port')}`;
    retval += `/${C.get('WOV_api_ver')}${_ep}`;
    return retval;
  }
};
