
const uuidv4   = require('uuid/v4');
const Logger   = require('woveon-logger');
const DB       = require('./db');

let logger    = new Logger('res', {debug : true, showName : true, level : 'verbose', color : 'blue'});


/**
 * The dual hiearchy builds the following fields:
 *
 * wovschema - The data sent to Woveon.
 * plschema  - The data stored in this Plugin database for the Res.
 * wovreq    - Required fields for this Res to be considered "complete".
 * wovdata   - All the data on the Res.
 * rsdata    - The Remote Server data that was sent.
 *
 *
 * Ids:
 *
 * inc
 * remoteId
 * xid
 */


/** Stores interface and schema for WovProto and RSProto. */
class GenericProto { /** */ constructor() { this.interface = {}; this.plschema = {}; this.wovschema = {}; }};

/** */
class WovProto extends GenericProto {

  /**
   * @param {string} _restype - the type of resource of this (ex. 'messages')
   * @param {array} _wovreq - required attributes of _wovdata
   */
  constructor(_restype, _wovreq) {
    super();

    // What needs to be set before it is 'complete'. append each data attribute to this in hierarchy
    this._wovreq = ['xid', 'remoteId', 'resInc'].concat(_wovreq);
    /*
    this._wovreq = Object.assign({}, {
      xid      : 'getXID',
      remoteId : 'getRemoteID',
      resInc   : 'getResInc',
    }, _wovreq);
    */

    this.restype = _restype;

    Object.assign(this.plschema, {
      xid      : String, // {type : DB.Mongoose.Types.UUID, required : false},
      resInc   : {type : DB.Mongoose.Types.Long, require : true, unique : true}, // unique to messages in plugin channel
      remoteId : String, // used to reply to this message
    });

    Object.assign(this.wovschema, {
      xid      : String, // {type : DB.Mongoose.Types.UUID, required : false},
      resInc   : {type : DB.Mongoose.Types.Long, require : true, unique : true}, // unique to messages in plugin channel
      remoteId : String, // used to reply to this message
    });
  }

  /**
   * An incrementing id for the Res.
   *
   * - While considered part of Woveon's data, it is set by the remote service's data by the RL.
   *
   * - IMPORTANT: This is unique ONLY for the Channel and Resource type.
   *    ie. no other Message from on a channel will have that resInc, but another channel could, or another resouce
   *        that is of type Person could as well.
   *
   * @return {Long} - the Woveon identifier of the resource.
   */
  getResInc()       { return this._wovdata.resInc; };

  /**
   * The unique id used by the Remote Service.
   *
   * - While considered part of Woveon's data, it is set by the remote service's data by the RL.
   *
   * @return {String} - the RemoteService identifier of the resource.
   */
  getRemoteID()     { logger.info('getRemoteID'); return this._wovdata.remoteId; };

  /**
   * The unique id used by the Remote Service.
   *
   * - While considered part of Woveon's data, it is set by the remote service's data by the RL.
   *
   * @return {String} - the RemoteService identifier of the resource.
   */
  getRemoteParentID() { return this._wovdata.remoteParentId || null; };

  /**
   * @return {uuid} - the Woveon xid identifier of the message.
   */
  getXID()          { return this._wovdata.xid; };

  /**
   * @return {uuid} - the Woveon xid identifier for the top-most message of the chain.
   */
  getXIDTop()       { return this._wovdata.xidTop; };

  /**
   * @return {uuid} - the Woveon xid identifier for the parent message of this (i.e. this (child) message is a reply to it).
   */
  getXIDParent()    { return this._wovdata.xidParent; };

  /**
   * Returns data for Woveon. Specialized by inheritance.
   * @return {object} - object with url and data attributes
   */
  _reqWovData() {
    // logger.info('_reqWovData called in ', __filename);
    let retval = {xid : this.getXID()};
    return retval;
  }


};


/** */
class RSProto extends GenericProto {

  /**
   * @param {string} _resplugin - the plugin/remote service type (ex. 'email')
   */
  constructor(_resplugin) {
    super();
    this.resplugin = _resplugin;
  }


  /**
   * For the remote server, use additional data (passed in or looked up) to obtain
   * the data specific for this RS that will complete this Res. This may not 
   * complete it, but it passes back the values it can find now.
   *
   * NOTE: Remember to call super so each level in hierarchy will complete.
   *
   * @param {Object} _withThisRes - either null (meaning lookup values) or a Res to use
   * @return {Object} - values that will be applied with 'applyPostProcessing'
   */
  async tryToCompleteWithRes(_withThisRes) { return {}; }


  /**
   * Returns the remote service/server's identifier from the passed in data.
   *
   * NOTE: static because Res and RL call this
   *
   * @param {object} _rawmsg - raw message data to parse if not null
   * @return {String} - A string is the most generic representation I could come up with.
   */
  static genRemoteID(_rawmsg = null) { throw new Error('child classes need to implement "genRemoteID" for the plugin'); return null; }


  /**
   * Generates the ordered increment of this message. Could be a timestamp (ex. ms since epoc) or
   * an actual order number (i.e. email IMAP ids). Must be a Long (64 bits).
   *    NOTE: there can be conflicts, if say this tracks milliseconds and different messages
   *    come in at the same time. The system handles this somewhat by checking RemoteIDs of messages.
   *   ex. return new Long(A, B);
   *
   *  NOTE: inc is unique for this channel
   *  NOTE: called by the RL as well.
   *
   * @param {object} _rawmsg - incoming message
   * @param {object} _lchannel - The mongo document for the local channel data.
   * @return {Long} - Mongoose.Types.Long
   */
  // static genResInc(_rawmsg, _lchannel) { throw new Error('child classes must implement "genResInc" for the plugin'); return null; }


  /**
   * Returns the data to store in the plugin.
   *
   * ex. For a message, data for replies and such.
   *
   * @return {object} - object with url and data attributes
   */
  _reqPLData() {
    let retval = {
      xid      : this.getXID(),
      resInc   : this.getResInc(),
      remoteId : this.getRemoteID(),
    };
    return retval;
  }


  /**
   * implement this for Remote services that need to create Res from the stored pldata
   * @param {object} _pldata -
   * @return {Res} -
   */
  /*
  generateResFromPLData(_pldata) {
    throw new Error('child classes must implement "generateResFromPLData" for the plugin');
    return null;
  }
  */

};


// ---------------------------------------------------------------------
//  ____           __  __           _      _
// |  _ \ ___  ___|  \/  | ___   __| | ___| |
// | |_) / _ \/ __| |\/| |/ _ \ / _` |/ _ \ |
// |  _ <  __/\__ \ |  | | (_) | (_| |  __/ |
// |_| \_\___||___/_|  |_|\___/ \__,_|\___|_|
// ---------------------------------------------------------------------

/** build a model, from WovProto and RSProto, that generates Res */
class ResModel {

  /**
   * @param {WovProto} _wovproto -
   * @param {RSProto} _rsproto -
   */
  constructor(_wovproto, _rsproto) {

    this.wovproto = _wovproto;
    this.rsproto  = _rsproto;
    this.models   = null;

    // Merge schema and requirements from two hierarchies
    this.plschema  = Object.assign({}, this.wovproto.plschema, this.rsproto.plschema);
    this.wovschema = Object.assign({}, this.wovproto.wovschema, this.rsproto.wovschema);
    // this.wovreq    = this.wovproto._wovreq.concat(this.rsproto._wovreq);

  }


  /**
   * Trace through and get all methods.
   *   https://stackoverflow.com/a/31055217/2664702
   * @param {object} _o -
   * @return {array} - names of methods
   */
  getProtoMethods(_o) {
    let retval = [];
    let obj = _o;

    // get methods on all classes (not base Object class)
    while (obj = Object.getPrototypeOf(obj)) {
      if ( obj.constructor.name != 'Object' ) retval = retval.concat(Object.getOwnPropertyNames(obj));
    }

    return retval.sort().filter(function(e, i, arr) {
       if (e!=arr[i+1] && (e != 'constructor' ) && (typeof _o[e]) == 'function') return true;
    });
  }

  /**
   * Merges methods from two different classes into one method, with one class taking priorty (i.e. if undefined, return the other). This
   * retains super() so functions can be inherited. This merges Wov and RS methods.
   *
   * NOTE: Remote Server is priority.
   *
   * @param {object} _t - this
   * @param {object} _n - function name
   * @return {function} - Looks for functions on rsproto, then wovproto to call. Returns, bound to 'this'.
   */
  mergeAF(_t, _n) {
    return (async function(...theArgs) {
      let ta = theArgs;
      let f1 = this._resmodel.rsproto[_n];
      let f2 = this._resmodel.wovproto[_n];
      let retval = undefined;
      if ( f1 != undefined ) { retval = f1.apply(this, ta); }
      if ( retval == undefined && f2 != undefined ) { retval = f2.apply(this, ta); }
      return retval;
    }).bind(_t);
  }

  /**
   * @param {object} _t - this
   * @param {object} _n - function name
   * @return {function} - Looks for functions on rsproto, then wovproto to call. Returns, bound to 'this'.
   */
  mergeF(_t, _n) {
    return (function(...theArgs) {
      let ta = theArgs;
      let retval = undefined;
      let f1 = this._resmodel.rsproto[_n];
      let f2 = this._resmodel.wovproto[_n];
      if ( f1 != undefined ) { retval = f1.apply(this, ta); }
      if ( retval == undefined && f2 != undefined ) { retval = f2.apply(this, ta); }
      return retval;
    }).bind(_t);
  }

  /**
  */
  displayMethods() {
    console.log(`ResModel: class ${this.constructor.name} wovproto(${this.wovproto.constructor.name}) rsproto(${this.rsproto.constructor.name})`);

    Object.getOwnPropertyNames(this.wovproto.interface)
      .forEach(( m ) => { console.log('  : (wovproto) ', m); });

    Object.getOwnPropertyNames(this.rsproto.interface)
      .forEach(( m ) => { console.log('  : (rsproto)  ', m); });
  }


  /**
   * Returns the name of the type of this resource. ex. 'messages'
   * @return {string} - the _name
   */
  getType() {
    let retval = this.wovproto.restype;
    // console.log('ResModel getType: ', this.wovproto.restype, retval);
    return retval;
  }


  /**
   * Returns the plugin type of this resource. ex. 'email'
   * @return {string} - the _name
   */
  getPlugin() { return this.rsproto.resplugin; }


  /**
   * Call when the database is accessible (i.e. in onInit methods).
   * ex. onInit() { ... EmailMessageModel.connect(_db); ... }
   * @param {DB} _db - mongoose database
   */
  connect(_db) {
    // create the connection to the model
    // NOTE: try to grab cached model, since Mongoose caches them inernally, and not per connection
    try { this.models = _db.connection.model(this.getType()); }
    catch (err) { // this is not an error, just a 'model is not cached so create it here'
      // logger.info(`PL Mongoose Schema for "${this.getType()}": `, Object.keys(plschema));
      this.models = _db.connection.model(this.getType(), this.plschema);
    }
  }


  /**
   * Uses the query to find an entry, returning a Res from this.
   * @param {object} _q - a mongodb query for findOne
   * @param {Res} - a Res or null if the query fails
   */
  async generateResFromQuery(_q) {
    let retval = null;
    let data   = await this.models.findOne(_q).lean();
    // logger.info('generateResFromQUeyr: ', _q, data);
    if ( data != null ) {
      retval = this.generateRes(data, data, {errorOnWovdataSchema : false});
      //retval.displayData('generateResFromQuery');
    }
    return retval;
  };


  /**
   * @param {object} _wovdata - data passed from woveon to create a resource
   * @param {object} _rsdata - data passed from a plugin to create a resource
   * @param {object} _options - additional options to pass
   * @return {Res} - Generates a Res object to hold that data
   */
  generateRes(_wovdata = null, _rsdata = null, _options = {}) {

    let options = Object.assign({errorOnWovdataSchema : true}, _options);

    // TODO make this generateRes and get rid of generateResWL and generateResRL

    // wovdata - this is a layering of wovreq, and _wovdata filtered through schema
    let wovdata = {};
    for (let i=0; i<this.wovproto._wovreq.length; i++) { wovdata[this.wovproto._wovreq[i]] = undefined; }
    for (let k in _wovdata) {
      if ( _wovdata.hasOwnProperty(k) ) {
        // logger.info('generateRes k: ', k);
        if ( this.wovschema[k] != undefined || this.plschema[k] != undefined ) { wovdata[k] = _wovdata[k]; }
        else { if ( options.errorOnWovdataSchema ) logger.throwError(`Res creation: passed non-schema validated field : ${this.getType()}::${k} `); }
      }
    }


    // create the Res object
    // logger.info('...building res from ', wovdata, _rsdata);
    let retval = new Res(this, wovdata, _rsdata);

    // add these in and overwrite if new ones in interface
    retval.getType   = (function() { return this._resmodel.getType(); }).bind(retval);
    retval.getPlugin = (function() { return this._resmodel.getPlugin(); }).bind(retval);


    // bind each function in wovproto and rsproto
    let wovprotomethods = this.getProtoMethods(this.wovproto); // NOTE: could cache this
    let rsprotomethods = this.getProtoMethods(this.rsproto);   // NOTE: could cache this
    for (let i=0; i<wovprotomethods.length; i++ ) {
      let f = wovprotomethods[i];
      if ( rsprotomethods.indexOf(f) == -1 ) {
        if ( f.constructor && f.constructor.name == 'AsyncFunction' ) { retval[f] = this.mergeAF(retval, f); }
        else { retval[f] = this.mergeF(retval, f); }
      }
    }
    for (let i=0; i<rsprotomethods.length; i++ ) {
      let f = rsprotomethods[i];
      if ( f.constructor && f.constructor.name == 'AsyncFunction' ) { retval[f] = this.mergeAF(retval, f); }
      else { retval[f] = this.mergeF(retval, f); }
    }

    // request WovData method
    retval.reqWovData = (function() {
      let retval = {
        url  : '/'+this.getType(),
        data : this._resmodel.wovproto._reqWovData.call(this),
      };
      return retval;
    }).bind(retval);

    // request PLData method
    retval.reqPLData = (function() {
      return {
        url  : '/'+this.getType(),
        data : this._resmodel.rsproto._reqPLData.call(this),
      };
    }).bind(retval);

    // generate xid for wovdata if not set
    retval._wovdata.xid = retval._wovdata.xid || uuidv4();

    return retval;
  }

};


// ---------------------------------------------------------------------
//  ____
// |  _ \ ___  ___
// | |_) / _ \/ __|
// |  _ <  __/\__ \
// |_| \_\___||___/
// ---------------------------------------------------------------------

/**
 * This Res object is specialized for a type (ex. Message) and plugin (ex. Email) through two
 * dual hierarchies of resource type (WovProto) and plugin resource type (RSProto). The methods on
 * this object are created by ResModel.generate, which is a compilation of both hierarchies.
 *
 * The dual hierarchies allow for a hierarchical tree of resource objects that Woveon can handle, that
 * are specialized for each different plugin of woveon (i.e. each remote service).
 *
 * Why? This seems complicated?
 * - Once understood, it creates a simplicity.
 *   - WovProto creats new types of Res.
 *   - RSProto specializes a Res for a specific Plugin/RemoteService.
 * - No changes inside code when:
 *   - as new types of data are added
 *   - as new Plugins/Remote Services are added
 *
 * So, to create an email message type, we just create a RSProto object that adds in schema and methods for email.
 */
class Res {

  /**
   * Just adds data. The resmodel, through its generateRes method, fills this with methods.
   * @param {ResModel} _resmodel - the object with all the methods this is bound to
   * @param {object} _wovdata    - data passed from woveon to create a resource
   * @param {object} _rsdata     - data passed from a plugin to create a resource
   */
  constructor(_resmodel, _wovdata, _rsdata) {
    Res.staticcounter = (Res.staticcounter+1) || 1;
    this.staticid  = Res.staticcounter;

    this._resmodel = _resmodel;
    this._wovdata  = _wovdata;
    this._rsdata   = _rsdata;


    // a Mongo Document... created when needed
    this._mymongodoc = null;

  }


  /**
   * Sets Woveon data (i.e. this._wovdata array) handed back from tryToCompleteWithRes.
   * @param {object} _wovdata - data to process with _postProcess
   */
  applyPostProcessing(_wovdata) {

    // logger.info('applyPostProcessing :', this._resmodel.wovproto._wovreq, _wovdata);
    for (let k in _wovdata) {
      if ( _wovdata.hasOwnProperty(k) ) {
        if ( this._resmodel.wovproto._wovreq.indexOf(k) != -1 ) {
          this._wovdata[k] = _wovdata[k];
          delete(_wovdata[k]);
        }
        else { logger.throwError(`trying to set non _wovdata value: {${k} : ${_wovdata[k]}}`); }
      }
    }

    // verify remaining dat is empty
    for (let k in _wovdata) {
      if (_wovdata.hasOwnProperty(k)) {
        let val = _wovdata[k];
        logger.throwError(`unconsumed data after postProcessing: {${k} : ${val}}`);
      }
    }
  }


  /**
   * Returns the schema for Woveon variables for this Res.
   * @return {object} -
   */
  getWovSchema() { return this._resmodel.wovschema; }

  /**
   * Returns the schema for variables stored in the Plugin database for this Res.
   * @return {object} -
   */
  getPLSchema() { return this._resmodel.plschema; }

  /**
   * Helper function.
   * @return {MongooseModel} - where Res are stored
   */
  model() { return this._resmodel.models; }


  /**
   * This appends data to the resource, but does not save it.
   */
  appendRSData(_data) {
    Object.assign(this._rsdata, _data);
  }


  /**
   * Displays the data of _wovdata and _rsdata, and can pass in a string to print at the top to label this.
   * @param {string} _s - label to print for this display of data
   */
  displayData(_s = '') {
    logger.with('add_to_stacklvl', 1).info(`Res.displayData: type(${this.getType()}) id(${this.staticid}) : ${_s}`);
    console.log(' wovdata: ', this._wovdata);
    console.log(' rsdata : ', this._rsdata);
  };


  /**
   * Displays the required fields and schema for this Res.
   * @param {string} _s - label to print for this display of data
   */
  displaySchema(_s = '') {
    logger.with('add_to_stacklvl', 1).info(`Res.displaySchema : type(${this.getType()}) : ${_s}`);
    console.log(' wovreq   : ', this._resmodel.wovproto._wovreq);
    console.log(' wovschema: ', this._resmodel.wovschema);
    console.log(' plschema : ', this._resmodel.plschema);
  }


  /**
   * One-line identifying data.
   * @param {string} _s - label to print for this display of data
   */
  displayIDs(_s = '') {
    logger.with('add_to_stacklvl', 1).info(`Res: type(${this.getType()}) : xid(${this.getXID()}), Inc(${this.getResInc()}), remoteId(${this.getRemoteID()}), xidTop(${this.getXIDTop()}), xidParent(${this.getXIDParent()}), remoteParentID(${this.getRemoteParentID()}), ${(this.isStub()?'stub':'')}${(_s?` : ${_s}`:'')}`);
    // logger.with('add_to_stacklvl', 1).info(`    ${JSON.stringify(this._wovdata)}`);
  }


  /**
   * A helper function that returns wovdata ids. By default, wov and remote but those can be changed.
   *
   * @param {boolean} _wov    - return wov xid data
   * @param {boolean} _remote - return remote ids
   * @param {string}  _msg    - a message can be added... useful for debugging spew
   * @return {object} - ids that were requested
   * TODO: use the wovschema to do this
   */
  getIDs(_wov = true, _remote = true, _msg = null) {
    let retval = {};
    if ( _wov )    { retval.xid = this.getXID(); retval.xidTop = this.getXIDTop(); retval.xidParent = this.getXIDParent(); }
    if ( _remote ) { retval.remoteId = this.getRemoteID(); retval.remoteParentId = this.getRemoteParentID(); };
    if ( _msg )    { retval.msg = _msg; }
    return retval;
  };


  /**
  */
  displayStructure(_s = '') {
    logger.with('add_to_stacklvl', 1).info(`Res.displayStructure: type(${this.getType()}) : ${(_s?` : ${_s}`:'')}`);

    this._resmodel.getProtoMethods(this).forEach( (f) => {
      console.log(`  : ${f}`);
    });
    /*
    Object.getOwnPropertyNames(this)
      .forEach(( m ) => {
        if ( m.startsWith('get') ) {
          if ( this[m]._origProto != null ) {
            console.log(`  : ${m}`, this[m]._origProto.constructor.name);
          }
          else { console.log(`  : ${m}`, this._resmodel.constructor.name); }
        }
      });
      */
    console.log('wovdata: ', this._wovdata);
    console.log('rsdata : ', this._rsdata);
    console.log(this.reqWovData());
    console.log(this.reqPLData());
  }


  /**
   * Return a method on the WovProto hierarchy, bound to this object.
   *
   * NOTE: 'this' is a running Res object
   *
   * ex. in interface methods, a getMessage calls the wovproto first (getting from wovdata), then checks plugin data
   *     function getMessage() { let msg = this.wovproto('getMessage')(); if ( msg == null ) msg = ... ; return msg; }
   *
   * @param {string} _methodname - a string of the name of the method
   * @return {method} - a method from the WovProto object, bound to this.
   */
  resMethod(_methodname) {
    let retval = this._resmodel.wovproto.interface[_methodname];
    if ( retval != null ) retval = retval.bind(this);
    return retval;
  }


  /**
   * Override how resources are printed.
   * @return {string} -
   */
  toString() {
    let retval = JSON.stringify(this, function(k, v) { if (k=='_wovdata' || k=='_rsdata') return v; else return null; });
    return retval;
  }


  /**
   * Complete, means this Res is ready to send to the Woveon API (as far as Woveon data is concernerd).
   *
   * Incomplete means that the resource is not complete, and the plugin will fix it later. Data missing or something. It
   * will be stored in the plugin until sipper finally finds all necessary data.
   *
   * NOTE: If there is some other data relationship based upon the plugin (i.e. it has to fetch other data first) then
   *       override this.
   * @return {boolean} - true if basic values are set
   */
  isComplete() {
    let retval = true;

    // logger.info('isComplete wovreq', this._resmodel.wovproto._wovreq);
    for (let i=0; i<this._resmodel.wovproto._wovreq.length; i++) {
      let k = this._resmodel.wovproto._wovreq[i];
      // logger.info('iscomplete test: ', k, this._wovdata[k]);
      if ( this._wovdata[k] === undefined ) { retval = false; /* logger.info('missing: ', k); */ }
    }

    // if ( this._wovdata.xid && this._wovdata.xidTop && this._wovdata.xidParent !== undefined &&
    //   this._wovdata.resInc !== undefined && this._wovdata.remoteId && this._wovdata.remoteParentId !== undefined ) retval = true;

    return retval;
  }


  /**
   * Returns the name of the type of this resource. ex. 'messages'
   * @return {string} - the _name
   */
  getType() { throw new Error('should be added by ResModel'); return null; }


  /**
   * Returns the plugin type of this resource. ex. 'email'
   * @return {string} - the _name
   */
  getPlugin() { throw new Error('should be added by ResModel'); return null; }

  /**
   * Returns the url and data of the Res for Message(create) and Put(update) operations.
   *
   * NOTE: generated by ResModel.
   *
   * @return {object} - data
   */
  reqWovData() { throw new Error('Need to implement'); return null; }


  /**
   * Returns the url and data of the Res for Get and Delete operations.
   *
   * NOTE: generated by ResModel.
   * NOTE2: when used? FIXME?
   *
   * @return {object} - data
   */
  // reqWovID()  { throw new Error('Need to implement'); return null; }


  /**
   * Returns the data to store in the Plugin. This should be limited to
   * that which is necesary for replies and other operations.
   *
   * NOTE: generated by ResModel.
   *
   * @return {object} - data
   */
  reqPLData()  { throw new Error('Need to implement'); return null; }


  /**
   * Generates a Mongoose Model Document for the Plugin from Res object data.
   *
   * This uses the currently connected model and the data returns from
   * reqPLData(), to generate a Mongo Document. This can be used to
   * save the resource to the plugin via: `r.genDocument().save()`
   *
   * @return {Document} - a Mongoose Document that links to a Mongoose Document.
   * @throws error if connection to a Mongoose Database has not been established
   */
  async updateMongo() {
    if ( this._resmodel.models == null ) throw new Error('Not connected to Mongoose database. You need to call "connect(db)" first.');
    let data   = this.reqPLData().data;

    // find/create if needed
    if ( this._mymongodoc == null ) {
      // logger.info('  ... find/create: ', data.resInc, data);
      this._mymongodoc = await this._resmodel.models.findOneAndUpdate(
        {resInc : data.resInc},
        data,
        {upsert : true, new : true});
      // logger.info('  ... find/create post: ');
    }

    // update existing Mongo Document
    else {
      // logger.info('  ... update');
      await this._mymongodoc.update(data);
    }

    // logger.info('  ... updateMongo returning: ', this._mymongodoc);
    return this._mymongodoc;
  }


  /**
   * Utility function. Checks by xid.
   *
   * @param {Array} _array - the array to check
   * @return {int} -
   */
  findPosInArray(_array) {
    let retval = -1; // assume not in by default

    if ( _array != null ) {
      for (let i=0; (i< _array.length) && (retval < 0); i++) {
        if (_array[i]._wovdata.xid == this._wovdata.xid) retval = i;
      }
    }

    return retval;
  }


  /**
   * Insert _res by Inc. Modifies the _array. Utilty function.
   *
   * @param {Array} _array - the Inc SORTED array to check
   * @return {Array} - array after insert
   */
  insertIntoSortedArray(_array) {
    let retval = null;

    if ( _array == null || _array.length == 0 ) { retval = [this]; }
    else {
      let doinsert = true;
      let i=0;

      // move past Inc
      for (i=0; i< _array.length; i++) {

        if ( _array[i].getResInc().equals(this.getResInc()) ) { doinsert = false; break; }
        else if ( _array[i].getResInc().greaterThan(this.getResInc()) ) { break; }
      }

      if ( doinsert ) { _array.splice(i, 0, this); }
      retval = _array;
    }

    return retval;
  }


};


// Return Module
module.exports = {

  // these are used for creating new Resource types
  prot : {
    WovProto     : WovProto,
    RSProto      : RSProto,
    GenericProto : GenericProto,
    logger       : logger,                    // passed as a utility
    Types        : DB.Mongoose.Schema.Types,  // passed as a utility
    Long         : DB.Mongoose.Types.Long,    // passed as a utility
  },

  // Used by the developer
  ResModel : ResModel,
  Res      : Res,
  Long     : DB.Mongoose.Types.Long, // here for utility... since inc can be of type Long
};
