
const ResLib   = require('../reslib').prot;
const logger   = ResLib.logger;

/**
 * Prototype definition of a Message on a remote service. ex. Facebook post, Email post.
 */
class MessageWovProto extends ResLib.WovProto {


  /**
   * Example data from WL (created by core/models/Post::createPostByUser, with null fields removed in transmission)
   *
   * {
   *  "type": "createPost",
   *  "token": "dc0e503d-067c-46f9-bc18-bab11dd6d83c",
   *  "recipient": null,
   *  "subject": "",
   *  "message": "Hi",
   *  "images": [],
   *  "videos": [],
   *  "xid": "fba37c2c-6400-40bb-a9f1-20be511761c3",
   *  "xidTop": "2e5bc716-939f-42b9-bc36-41f117321dcf",
   *  "xidParent": "2e5bc716-939f-42b9-bc36-41f117321dcf"
   *}
   */

  /**
  */
  constructor() {
    super('posts', ['xidTop', 'xidParent', 'remoteParentId']);

    // Schema is for saved Plugin Data
    Object.assign(this.plschema, {
      resInc         : {type : ResLib.Long, require : true, unique : true}, // unique to messages in plugin channel
      xidParent      : String,       // {type : DB.Mongoose.Types.UUID, required : false},
      xidTop         : String,       // {type : DB.Mongoose.Types.UUID, required : false},
      remoteParentId : String,
      isstub         : Boolean,
    });

    // Schema for _wovdata
    Object.assign(this.wovschema, {
      resInc         : {type : ResLib.Long, require : true, unique : true}, // unique to messages in plugin channel
      xidParent      : String,       // {type : DB.Mongoose.Types.UUID, required : false},
      xidTop         : String,       // {type : DB.Mongoose.Types.UUID, required : false},
      remoteParentId : String,
      isstub         : Boolean,
      subject        : String,
      token          : String,       // channel token
      author         : String,
      message        : String,
    });
  };

  /**
   * @return {bool} - true if this has a parent, but the xidTop and xidParent are not pointing to them
   * because we have yet to find the parent. This message will have an entry in db.parentmissing.
   */
  isParentMissing() {
    if ( this._wovdata && this._wovdata.parentmissing === true ) return true;
    return false;
  };


  /**
   * Stub messages are placeholders for remote systems that don't track hierarchy (and Woveon does).
   * @return {boolean}
   */
  isStub() { return this._wovdata.isstub; };


  /**
   * @return {string} - The token for the channel the message is on.
   */
  getToken() { return this._wovdata.token || null; };


  /**
   * @return {string} - The creator of the message
   */
  getAuthor() {
    if (this._wovdata.author) { return this._wovdata.author; }
    return null;
  };


  /**
   * @return {string} - author's first name
   */
  getAuthorFirst() {
    if (this._rsdata && this._rsdata.author && this._rsdata.author.first) { return this._rsdata.author.first; }
    // else if (this._wovdata.author && this._wovdata.author != '') { return this._wovdata.author.split(' ')[0]; }
    else return null;
  };


  /**
   * @return {url} - url to the author's representation (ex. an icon)
   */
  getAuthorImg() {
    if (this._wovdata.authorUrl) { return this._wovdata.authorUrl || null; }
    else return null;
  };


  /**
   * @return {string} - the subject
   */
  getSubject() {
    if (this._wovdata.subject) { return this._wovdata.subject; }
    // else if (this._rsdata && this._rsdata.subject) { return this._rsdata.subject; }
    else return null;
  };


  /**
   * @return {string} - the message (mime type text for now)
   */
  getMessage() {
    if (this._wovdata.message) { return this._wovdata.message; }
    // TODO: convert _rsdata.message into a Cloud Storage URL safe message and set as this._wovdata.message and return that
    else return null;
  };


  /**
   * @return {array} - array of strings pointing to the location of the images
   */
  getImages() {
    if (this._wovdata.images) { return this._wovdata.images; }
    else return null;
  };


  /**
   * @return {array} - array of strings pointing to the location of the videos
   */
  getVideos() {
    if (this._wovdata.videos) { return this._wovdata.videos; }
    else return null;
  };


  /**
   * @return {string} - the person receiving the message
   */
  getRecipient() {
    if (this._wovdata.recipient) { return this._wovdata.recipient; }
    else return null;
  };


  /**
   * @return {Date} - i think a date object?
   */
  getDate() {
    if (this._wovdata.date) { return this._wovdata.date; }
    else return null;
  };


  /**
   * @return {object} - a data representation of this for Woveon.
   */
  _reqWovData() {
    let retval = super._reqWovData();

    // identifiers
    retval.xid            = this.getXID();
    retval.xidTop         = this.getXIDTop();
    retval.xidParent      = this.getXIDParent();

    // No need to send... plugin manages these
    retval.resInc         = this.getResInc();         // NOTE: used by Woveon to check for duplicates only
    // retval.remoteId       = this.getRemoteID();

    // Woveon data
    retval.token          = this.getToken();       // channel token
    retval.author         = this.getAuthor();      // author name
    retval.authorUrl      = this.getAuthorImg();   // image of author of message
    retval.subject        = this.getSubject();     // message subject
    retval.message        = this.getMessage();     // message body
    retval.images         = this.getImages();      // images in message
    retval.videos         = this.getVideos();      // videos in message
    retval.recipient      = this.getRecipient();   // target of message
    retval.date           = this.getDate();
    retval.rsraw          = this._rsdata;          // is this correct? store the raw version of the message
    // logger.info('reqWovData: ', retval);
    return retval;
  }


  /**
   * For most Messages, we just need to compute: xidTop, xidParent, remoteId, remoteParentId
   * NOTE: lack of a parent will leave these as incomplete, which is fine
   * @param {arrayOfRes} _wovres - array of newly created Res
   @return {object} - {wovres : array of new Res, wovresmods : existing Res that are modified}
   */
  /*
  async postProcessRes_Messages_Hierarchy(_wovres) {

    _wovres.forEach( (r) => {
      r.getRemoteID();
      r.getRemoteParentID();
      r.getXIDParent();
      /*
      r.getXIDTop();
      /*
    });

    return {wovres : _wovres, wovresmods : [] };
  }
  */


};


module.exports = MessageWovProto;
