
const RemoteDataSipper = require('../src/microservices/remoteDataSipper');

// NOTE: all time should be in plugin time so we don't worry about local/remote time issues.

/**
 * Used to download posts over time.
 */
module.exports = class DownloaderEngine extends RemoteDataSipper {

  /**
   * @param {Long} _monitor - monitoring start "time"
   * @param {Logger} _logger - logger
   */
  constructor(_monitor, _logger) {
    super({
      port         : 80,
      name         : 'DownloaderEngine',
      logger       : _logger,
      FirstMsgTest : RemoteDataSipper.FMT_LimitNotReached,
      sipsize      : 10,
    }, null);

    this.rs = null; // remote service
    this.resetSim(_monitor);
  }


  /**
   * @param {int} _monitor - new time to monitor from
   */
  resetSim(_monitor) {
    if ( typeof _monitor != 'object' ) { this.logger.throwError('passing non Long to resetSim ', _monitor); }
    this.posts = [];
    this.channel = { // simulated... same json as ChannelModel
      sipBlocks : [],
      e_mon     : _monitor,
    };
  }


  /**
   * remoteID is index 1, inc is index 0
   */
  genMsgRemoteID(_rawmsg) { if ( _rawmsg === undefined ) this.logger.throwError('undefined data'); return _rawmsg != null ? _rawmsg[1]: null; }


  /**
   * @param {object} _rawmsg - the remote server data
   * @param {object} _channel - unused
   * @return {object} - the rawmsg
   */
  genMsgInc(_rawmsg, _channel) { if ( _rawmsg === undefined ) this.logger.throwError('undefined data'); return _rawmsg != null ? _rawmsg[0]: null; }


  /**
   * Return server's Inc
   */
  async getInc() { return this.rs._curtime; }


  /**
   * @param {int} _limit - number of messages to grab
   * @return {integer} - number of data retreived
   */
  async doLoop(_limit) {
    let retval = 0;
    this.logger.info(`doLoop: ${_limit}`);
    let sipped = await this.sipChannelData(_limit, this.channel)
      .catch((err) => {
        this.logger.info(err);
        err.message.should.equal('Looks like you pruned all messages returned...' +
          ' inf loop incoming unless you increase limit size');
        return [];
      });

    let data = sipped.data;

    retval = data.length;

    // add in messages, store locally in this class

    // no data, easy
    if ( retval == 0 ) { this.logger.info(`  - doLoop added no data`); }

    // no posts so just dump in
    else if ( this.posts == null || this.posts.length == 0 ) {
      this.posts = data;
    }

    // add messages in, in order
    //
    // This is a PAIN IN THE A*S because we can't be assured that the Inc of messges are unique, since
    // some messages can have same Inc and we have to preserve order. These cases are extremely rare I would
    // imagine in most cases, but still could reorder the messages. This could happen in Facebook, but not
    // IMAP for instance because Facebook is ordered by time but IMAP has unique identifiers.
    //
    // Algorithm to insert messages in order:
    // * if we pruned, then use earliest pruned message, else latest msg
    // * start at retreived messages, at most recent
    //   * if ids match to pruned, put before that msg
    //   * if inc is less thatn data, put before that msg
    else {
      let i_inc = null;
      let i_id  = null;
      this.logger.info('sip returned ', sipped);
      if ( sipped.pruned.length != 0 ) {
        i_inc = this.genMsgInc(sipped.pruned[0] || null, this.channel);
        i_id  = this.genMsgRemoteID(sipped.pruned[0]||null);
        this.logger.info(`  ... i_inc: ${i_inc}   i_id: ${i_id}`);
      }
      else {
        i_inc = this.genMsgInc(sipped.data[sipped.data.length-1] || null, this.channel);
        this.logger.info(`  ... i_inc: ${i_inc}`);
      }

      // can only know if less or same/greater
      let cmpFunc = function(i_inc, i_id, p_inc, p_id) {
        let retval = -1;
        this.logger.info(`  ... cmpFunc ${i_inc},${i_id}  ${p_inc},${p_id}`);


        if ( i_inc.greaterThan(p_inc) ) {
          this.logger.info('  ...clearly larger');
          retval = 1;
        }

        // if inc matches and we hav a prune with the id, insert before
        else if (i_id != null && i_inc.equals(p_inc) ) {
          if (  i_id == p_id) {
            this.logger.info(`  ... put data before post, since pruned is the post`);
            retval = 0;
          }
          else {
            this.logger.info(`  ... put data after post, since we assume going right to left`);
            retval = -1;
          }
        }

        // else (no prunes) then just insert before inc
        else if ( i_inc.equals(p_inc) ) {
          this.logger.info('  ...put data before psot');
          retval = 0;
        }
        this.logger.info(`  ... cmpFunc ${retval}`);
        return retval;
      };

      let i = this.posts.length - 1;
      let p_id  = this.genMsgRemoteID(this.posts[i] || null);
      let p_inc = this.genMsgInc(this.posts[i] || null, this.channel);
      this.displayMe('pre add in data');

      // if equal or greater, then append
      if ( cmpFunc.call(this, i_inc, i_id, p_inc, p_id) == 1 ) {
        this.logger.info(`  ... splice in at end... not sure if works ${i}`);
        [].splice.apply(this.posts, [this.posts.length, 0].concat(sipped.data)); // javascript syntactic dark magic
        this.displayMe('post add in data');
//        throw Error('check values');
      }

      // else, find spot to splice in (which might be at start so don't check i==0)
      else {
        for (; i>0; i--) {
          p_id  = this.genMsgRemoteID(this.posts[i] || null);
          p_inc = this.genMsgInc(this.posts[i] || null, this.channel);
          this.logger.info(`  ... check pos ${i}:    p_id: ${p_id}  p_inc: ${p_inc}`);
          if ( cmpFunc.call(this, i_inc, i_id, p_inc, p_id) >= 0 ) { break; }
        }

        this.logger.info(`  ... work up posts, checking ids if i!= 0    i:${i}`);
        if ( i > 0 ) {
          // let ltrim = i;
          let p_inc = this.genMsgInc(this.posts[i-1], this.channel);
          let p_id  = this.genMsgRemoteID(this.posts[i-1]);
          let cmp = null;
          sipped.lpruned = [];
          do {
            let j_inc = this.genMsgInc(sipped.data[0] || null, this.channel);
            let j_id  = this.genMsgRemoteID(sipped.data[0] || null);
            cmp = cmpFunc.call(this, j_inc, j_id, p_inc, p_id);
            this.logger.info(`  ...cmp returned ${cmp}`);
            if ( cmp <= 0 ) {
              this.logger.info('  ...pop off sipped msg');
              let m = sipped.data.shift();
              sipped.lpruned.push(m);
            }
          } while (cmp <= 0 && sipped.data.length > 0 );
        }

        this.logger.info(`  ... prune inserting data at post position ${i}`);
        [].splice.apply(this.posts, [i, 0].concat(sipped.data)); // javascript syntactic dark magic
      }

      this.displayMe('post add in data');

//      this.logger.throwError('hiere');
    }

    return retval;
  }


  /**
   * The service calls this to simulate the posting of data to this listening engine. handleNotifiedMessages removes overlaps.
   * @param {*} _updatetime -
   * @param {*} _data -
   */
  async _addMsg(_updatetime, _data) {
    this.logger.info(`_addMsg: ${_updatetime} ${_data}`);
    this.displayMe('before addding post');
    let allvalid = await this.handleNotifiedMessages( this.channel, [[_updatetime, _data]], false);
//    let isvalid = this.isValidRSNotification([_updatetime,_data],this.channel,this.rs,this.logger)
    this.logger.info(' handle ', allvalid);
    if ( allvalid ) {
      this.posts.push([_updatetime, _data]);
      this.posts.sort(DownloaderEngine._compareEntries);
      this.displayMe('after addding post');
    }
    else { this.logger.info(`_addMsg skipping post ${_updatetime}, ${_data}`); }
  }


  /**
   * From this time, back, up to limit.
   * @param {*} _time  - start fetching from BELOW here and back (not inclusive)
   * @param {*} _limit  - number to get
   * @param {object} _channel -
   * @return {array} - of posts, in increasing time
   */
  async fetchOnAndBefore(_time, _limit, _channel) {

    if ( _time == 'now' ) _time = this.rs._curtime;
    this.logger.info('_time: ', _time);
    // console.trace();

    let i = -1;
    let bounds = [-1, -1]; // [start,end] where start > end
    let retval = {curinc : null, data : [] };

    this.logger.aspect('fetchOnAndBefore', `fetchOnAndBefore : t:${_time} limit:${_limit}`);

    // find start, starting at end
    i = this.rs.posts.length-1;
    while (i >= 0 && this.rs.posts[i][0].greaterThan(_time)) { i--; } // OnAnd
    bounds[1] = i;

    // find end by counting down and then making sure bigger than 0
    i = i-_limit+1;
    i = i<0? 0:i;
    bounds[0] = i;
    this.logger.aspect('fetchOnAndBefore', `  - : ${bounds[0]} <-> ${bounds[1]} : inclusive index to return`);

    // gather those (ordered by default since posts are ordered)
    for (i=bounds[0]; i<=bounds[1]; i++) {
      retval.data.push( [this.rs.posts[i][0], this.rs.posts[i][1]] );
    }
    this.logger.aspect('fetchOnAndBefore', `  - : ${JSON.stringify(retval)}`);


    // ** determine where we are in the fetching of messages, set curinc to that
    // if we have data, get Inc of the earliest message
    if ( retval.data.length > 0 && retval.data.length == _limit ) { retval.curinc = this.genMsgInc(retval.data[0], _channel); }

    // if limit is 0, then ignore
    else if ( _limit == 0 ) { retval.curinc = _time; }

    // else, no data but we tried, done : NOTE: not all RemoteServices have this assumption, but this testing service does
    else {
      if ( this.channel.sipBlocks.length > 0 ) { retval.curinc = this.channel.sipBlocks[0].f; } // set to first
      else retval.curinc = null;
    }

    /*
    retval.curinc = (retval.data.length > 0 ? this.genMsgInc(retval.data[0], _channel): null);
    if ( retval.curinc == null ) retval.curinc = _time;
    */

    this.logger.info('returning', retval);
    return retval;
  }


  /**
   * Comparison function of the Inc values (Long).
   * @param {Array} a - Long value in position 0
   * @param {Array} b - Long value in position 0
   * @return {int} - 1 >, 0 ==, -1 <
   */
  static _compareEntries(a, b) {
    if ( a[0].lessThan(b[0]) ) return -1;
    else if ( a[0].greaterThan(b[0]) ) return 1;
    else return 0;
  }

  /**
   * @param {object} _msg -
   */
  displayMe(_msg) {
    this.logger.with('add_to_stacklvl', 1).info(`${this.constructor.name}: #:${this.posts.length} (M:${this.channel.e_mon})  : ${_msg || ''}`);
    for (let i=0; i< this.channel.sipBlocks.length; i++) {
      // let b = this.channel.sipBlocks[i];
      this.logger.info(`  - sipBlock ${i} : (f:${this.channel.sipBlocks[i].f} --`+
                       ` p:${this.channel.sipBlocks[i].p} -- l:${this.channel.sipBlocks[i].l})`+
                       ` ${this.channel.sipBlocks[i].finished}`);
    }
    for (let i=0; i< this.posts.length; i++) {
      let e = this.posts[i];
      this.logger.info(`${i}: '${e}'`);
    }
  }

};
