
const should            = require('should');
const nock              = require('nock');
const urlParse          = require('url-parse');

const Logger            = require('woveon-logger');

const config            = require('../service/config');
const Service           = require('../service/service');
const RemoteListenerService     = require('../microservices/remoteListenerService');
const WoveonListenerService = require('../microservices/woveonListenerService');

const Helper            = require('./helper');
const dbHelper          = require('./dbHelper');


module.exports = () => {

  let mtag = require('path').basename(__filename).slice(0,-3);
  let logger = new Logger(mtag,{debug: true, showName: true});
  let wconfig = JSON.parse(JSON.stringify(config));
  let gconfig = JSON.parse(JSON.stringify(config));
  wconfig.service.port = process.env.WOV_ple_rl_port;
  gconfig.service.port = config.plugin.wl_port;

  let wovonelistener = new WoveonListenerService(gconfig, {
    logger     : new Logger('G', {showName: true, debug: true},
      { listener: false, requester: false }), });

  let remotelistener = new RemoteListenerService(wconfig, { 
    logger: new Logger('W', { showName: true, debug: true, level: 'verbose' },
        { listener: true, requester: false, woveon: true }) });

  // inited values
  let session      = null;
  let channelToken = null;

  describe(`${mtag}: Woveon SDK testing`, function() {
    logger.h1().info(this.title);


    // Create testing microservices for Facebook plugin
    before( async () => {
      await Helper.modelsClear();
      let pg = await dbHelper.getWoveonDBConnection();
      await dbHelper.clearTables(pg); // posts, comments, likes, teamchannel
      await pg.end();
      await remotelistener.init();
      await remotelistener.startup()
        .catch((err)=>{console.log(err); throw new Error('failed remotelistener.startup');});
      await wovonelistener.init()
        .catch((err)=>{console.log(err); throw new Error('failed wovonelistener.init');});
      await wovonelistener.startup()
        .catch((err)=>{console.log(err); throw new Error('failed wovonelistener.startup');});
      session = await Helper.initAPISession(wovonelistener)
        .catch((err)=>{console.log(err); throw new Error('failed initAPISession');});
      channelToken = await Helper.createNewChannel(wovonelistener, session)
        .catch((err)=>{console.log(err); throw new Error('failed createNewChannel');});
    });

    describe(`${mtag}: A - Posts`, function() {
      logger.h1().info(this.title);

      let post_orig = null;
      let post_created = null;
      let post_update = null;
      let post_updated = null;
      let post_delete = null;
      let post_deleted = null;

      it(`${mtag}: A.1 - createPublicPost`, async function() {
        logger.h2().info(this.test.title);

        post_orig = {
          token: channelToken,
          author: 'Sam Spade',
          authorUrl: 'https://moviewriternyu.files.wordpress.com/2013/08/spade.jpg',
          subject: 'A Subject',
          isDirect: false,
          postId: Service.generateRandomString(10),
          message: 'A Message',
          images: [],
          videos: [],
        };
        let result = await remotelistener.toWoveon.createPost(post_orig);
        logger.info('createPost result: ', result);
        result.success.should.be.true();
        should.notEqual(result.data, null);

        // added values
        should.notEqual(result.data.teamchannelid, null);
        should.notEqual(result.data.id, null);

        // echoed values
        result.data.author.should.equal(post_orig.author);
        result.data.authorUrl.should.equal(post_orig.authorUrl);
        result.data.subject.should.equal(post_orig.subject);
        result.data.isDirect.should.equal(post_orig.isDirect);
        result.data.postId.should.equal(post_orig.postId);
        result.data.message.should.equal(post_orig.message);
        result.data.images.length.should.equal(post_orig.images.length);
        result.data.videos.length.should.equal(post_orig.videos.length);

        post_created = result.data;
      });

      it(`${mtag}: A.1.1 - createPrivatePost`);
      it(`${mtag}: A.1.2 - createPost`);


      it(`${mtag}: A.2 - updatePost`, async function() {
        logger.h2().info(this.test.title);

        // should catch bad updates
        try {
          let fail = await remotelistener.toWoveon.updatePost({});
          throw new Error("XXX-Should have failed with empty update.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch upate'); }

        try {
          let fail = await remotelistener.toWoveon.updatePost({ token: 'token' });
          throw new Error("XXX-Should have failed with missing id.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch upate'); }

        try {
          let fail = await remotelistener.toWoveon.updatePost({ id: 'id' });
          throw new Error("XXX-Should have failed with missing token.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch upate'); }

        post_update = {
          token: post_orig.token,
          id: post_created.id,
          message: 'A new message',
        };
        logger.info('post_update: ', post_update);

        let result = await remotelistener.toWoveon.updatePost(post_update);
        logger.info('post_update: ', post_update);
        logger.info('updatePost result: ', result);
        result.success.should.be.true();
        result.data.teamchannelid.should.equal(post_created.teamchannelid);
        logger.info(` ${result.data.id} ${post_update.id}`);
        result.data.id.should.equal(post_update.id);
        result.data.message.should.equal(post_update.message);

        post_updated = result.data;
      });




      it(`${mtag}: A.3 - deletePost`, async function() {
        logger.h2().info(this.test.title);

        // should catch bad updates
        try {
          let fail = await remotelistener.toWoveon.deletePost({});
          throw new Error("XXX-Should have failed with empty delete.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch delete'); }

        try {
          let fail = await remotelistener.toWoveon.deletePost({ token: 'token' });
          throw new Error("XXX-Should have failed with missing id.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch delete'); }

        try {
          let fail = await remotelistener.toWoveon.deletePost({ id: 'id' });
          throw new Error("XXX-Should have failed with missing token.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch delete'); }

        post_delete = {
          token: post_orig.token,
          id: post_created.id,
        };
        logger.info('post_delete: ', post_delete);

        let result = await remotelistener.toWoveon.deletePost(post_delete);
        logger.info('post_delete: ', post_delete);
        logger.info('deletePost result: ', result);
        result.success.should.be.true();
        should.equal(result.data, null);

        post_deleted = result.data;
      });
    });


    describe(`${mtag}: B - Comments`, function() {
      logger.h1().info(this.title);

      let post_orig       = null;
      let post_created    = null;
      let comment_orig    = null;
      let comment_created = null;
      let comment_update  = null;
      let comment_updated = null;


      before(async () => {
        console.log(`${mtag}: B.Before - create post for comments`);
        post_orig = {
          token: channelToken,
          author: 'Sam Spade',
          authorUrl: 'https://moviewriternyu.files.wordpress.com/2013/08/spade.jpg',
          subject: 'A Subject',
          isDirect: false,
          postId: Service.generateRandomString(10),
          message: 'This is a post.',
          images: [],
          videos: [],
        };
        let result = await remotelistener.toWoveon.createPost(post_orig);
        logger.info('create Post result: ', result);
        result.success.should.be.true();
        post_created = result.data;
      });


      it(`${mtag}: B.1 - createComment`, async function() {
        logger.h2().info(this.test.title);

        comment_orig = {
          token     : channelToken,
          author    : 'Sam Spade2',
          authorUrl : 'https://moviewriternyu.files.wordpress.com/2013/08/spade2.jpg',
          postId    : post_created.postId,
          commentId : `${post_created.postId}_${Service.generateRandomString(10)}`,
          message   : 'Comment message',
          images    : [],
          videos    : [],
        };
        let result = await remotelistener.toWoveon.createComment(
          comment_orig.token, comment_orig.author, comment_orig.authorUrl,
          comment_orig.postId, comment_orig.commentId, comment_orig.message);
        logger.info('createComment result: ', result);
        result.success.should.be.true();
        should.notEqual(result.data, null);

        // added values
        should.notEqual(result.data.id, null);
        result.data.postid.should.equal(post_created.id);


        // echoed values
        result.data.author.should.equal(comment_orig.author);
        result.data.authorUrl.should.equal(comment_orig.authorUrl);
        result.data.message.should.equal(comment_orig.message);
        result.data.commentId.should.equal(comment_orig.commentId);
        should.equal(result.data.images,null); // null for some reason...
//        result.data.videos.length.should.equal(post_orig.videos.length); // video is a {}

        // deleted
        should.equal(result.data.postId, null);

        comment_created = result.data;

      });


      it(`${mtag}: B.2 - updateComment`, async function() {
        logger.h2().info(this.test.title);

        // should catch bad updates
        try {
          let fail = await remotelistener.toWoveon.updateComment({});
          throw new Error("XXX-Should have failed with empty update.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch upate'); }

        try {
          let fail = await remotelistener.toWoveon.updateComment({ token: 'token' });
          throw new Error("XXX-Should have failed with missing id.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch upate'); }

        try {
          let fail = await remotelistener.toWoveon.updateComment({ commentId: 'id' });
          throw new Error("XXX-Should have failed with missing token.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch upate'); }

        comment_update = {
          token: comment_orig.token,
          commentId: comment_created.commentId,
          message: 'An updated comment',
        };
        logger.info('comment_update: ', comment_update);

        let result = await remotelistener.toWoveon.updateComment(comment_update);
        logger.info('comment_update: ', comment_update);
        logger.info('updateComment result: ', result);

        result.success.should.be.true();
        should.notEqual(result.data,null);
        should.notEqual(result.data.date,null);

        result.data.postid.should.equal(post_created.id);
        result.data.author.should.equal(comment_created.author);
        result.data.authorUrl.should.equal(comment_created.authorUrl);
        result.data.message.should.equal(comment_update.message);
        result.data.commentId.should.equal(comment_created.commentId);
        should.equal(result.data.images,null); // null for some reason...
        //should.equal(result.data.video,{}); // {} for some reason...

        comment_updated = result.data;
      });


      it(`${mtag}: B.3 - deleteComment`, async function() {
        logger.h2().info(this.test.title);

        // should catch bad updates
        try {
          let fail = await remotelistener.toWoveon.deleteComment({});
          throw new Error("XXX-Should have failed with empty delete.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch delete'); }

        try {
          let fail = await remotelistener.toWoveon.deleteComment({ token: 'token' });
          throw new Error("XXX-Should have failed with missing id.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch delete'); }

        try {
          let fail = await remotelistener.toWoveon.deleteComment({ id: 'id' });
          throw new Error("XXX-Should have failed with missing token.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch delete'); }

        comment_delete = {
          token: comment_orig.token,
          id: comment_created.id,
        };
        logger.info('comment_delete: ', comment_delete);

        let result = await remotelistener.toWoveon.deleteComment(comment_delete);
        logger.info('comment_delete: ', comment_delete);
        logger.info('deleteComment result: ', result);
        result.success.should.be.true();
        should.equal(result.data, null);

        comment_deleted = result.data;
      });

    });


    describe(`${mtag}: C - Like`, function() {
      logger.h1().info(this.title);

      let post_orig       = null;
      let post_created    = null;
      let comment_orig    = null;
      let comment_created = null;
      let like_orig       = null;
      let like_created    = null;


      before(async () => {
        console.log(`${mtag}: C.Before - create post and comment for like/unlike`);
        post_orig = {
          token: channelToken,
          author: 'Sam Spade3',
          authorUrl: 'https://moviewriternyu.files.wordpress.com/2013/08/spade3.jpg',
          subject: 'A Subject 3',
          isDirect: false,
          postId: Service.generateRandomString(10),
          message: 'This is a post 3.',
          images: [],
          videos: [],
        };
        let result = await remotelistener.toWoveon.createPost(post_orig);
        logger.info('create Post result: ', result);
        result.success.should.be.true();
        post_created = result.data;


        comment_orig = {
          token     : channelToken,
          author    : 'Sam Spade4',
          authorUrl : 'https://moviewriternyu.files.wordpress.com/2013/08/spade4.jpg',
          postId    : post_created.postId,
          commentId : `${post_created.postId}_${Service.generateRandomString(10)}`,
          message   : 'Comment message 4',
          images    : [],
          videos    : [],
        };
        result = await remotelistener.toWoveon.createComment(
          comment_orig.token, comment_orig.author, comment_orig.authorUrl,
          comment_orig.postId, comment_orig.commentId, comment_orig.message);
        logger.info('createComment result: ', result);
        result.success.should.be.true();
        comment_created = result.data;
      });


      it(`${mtag}: C.1 - doLike post`, async function() {
        logger.h2().info(this.test.title);

        like_orig = {
          token     : post_orig.token,
          postId    : post_orig.postId, 
          //id        : post_created.id, 
          author    : 'Sam Spade5',
          authorUrl : 'https://moviewriternyu.files.wordpress.com/2013/08/spade5.jpg',
        };
        logger.info('like_orig: ', like_orig);
        let result = await remotelistener.toWoveon.doLike(like_orig);
        logger.info('doLike result: ', result);
        result.success.should.be.true();
        should.notEqual(result.data,null);

        result.data.id.should.be.equal(post_created.id);
        should.notEqual(result.data.date,null);
        result.data.postId.should.be.equal(post_orig.postId);
        result.data.message.should.be.equal(post_orig.message);

        like_created = result.data;
      });


      it(`${mtag}: C.1.1 - multiple doLike post`);


      it(`${mtag}: C.2 - doLike comment`, async function() {
        logger.h2().info(this.test.title);

        likec_orig = {
          token     : post_orig.token,
          postId    : post_orig.postId, 
          commentId : comment_orig.commentId,
          author    : 'Sam Spade6',
          authorUrl : 'https://moviewriternyu.files.wordpress.com/2013/08/spade6.jpg',
        };
        logger.info('likec_orig: ', likec_orig);

        try {
          let fail = await remotelistener.toWoveon.doLike(likec_orig);
          throw new Error("XXX-Should have failed with missing token.");
        } catch (err) { if (err.message.startsWith('XXX-')) throw new Error('Failed to catch doLike bad args'); }

        delete likec_orig.postId;
        let result = await remotelistener.toWoveon.doLike(likec_orig);
        logger.info('doLike (c) result: ', result);
        result.success.should.be.true();
        should.notEqual(result.data,null);

        result.data.id.should.be.equal(comment_created.id);
        should.notEqual(result.data.date,null);
        result.data.message.should.be.equal(comment_orig.message);

        likec_created = result.data;
      });


      it(`${mtag}: C.3 - undoPostLike`,
      /* async () => {
        let obj = {
          token  : like_orig.token, 
          id     : post_created.id,
          author : like_orig.author,
        };

//        API can't find post from the given id... not sure what is up
        logger.info('undoPostLike : ', obj);
        let result = await remotelistener.toWoveon.undoPostLike(obj);
        logger.info('undoPostLike result: ', result);
        result.success.should.be.true();
        should.notEqual(result.data,null);
        result.data.id.should.be.equal(post_created.id);
        result.data.date.should.be.equal(post_created.date);
        result.data.message.should.be.equal(post_created.message);
        result.data.postId.should.be.equal(post_created.postId);


        obj.id = likec_created.id;
        obj.author = likec_orig.author;
        logger.info('undoLike (c) : ', obj);
        result = await remotelistener.toWoveon.undoCommentLike(obj);
        logger.info('undoLike (c) result: ', result);
        result.success.should.be.true();
        should.notEqual(result.data,null);
        result.data.id.should.be.equal(comment_created.id);
        result.data.date.should.be.equal(comment_created.date);
        result.data.message.should.be.equal(comment_created.message);

      }*/
    );


      it(`${mtag}: C.3.1 - multiple undoLike`);
    });


    after(async () => {
      logger.info('\n\n************************************************************');
      logger.info('logger ending');
      await remotelistener.shutdown();
      await wovonelistener.shutdown();
      await Helper.modelsClear();
    });

  });

}