
// const mongoose         = require('mongoose');
const bcrypt           = require('bcryptjs');
const URL              = require('url-parse');
const qs               = require('querystringify');
const nock             = require('nock');
const autoBind         = require('auto-bind-inheritance');


const Logger            = require('woveon-logger');

const config           = require('../service/config');
const DB               = require('../service/db');
const Requester        = require('../service/requester');
const ChannelModel     = require('../models/channelModel');
const RemoteListenerModel = require('../models/remoteListenerModel');
const WovMsgsModel     = require('../models/wovMsgsModel');
const DownloaderEngine = require('./downloaderEngine');
const RemoteListenerService    = require('../microservices/remoteListenerService');

const SALT_WORK_FACTOR = 10;

let logger    = new Logger('h',{debug: true, showName: true, level: 'info'});
let requester = new Requester(logger);
let apiUrl    = new URL(config.api.url); 
  // NOTE: apiUrl does not use the ap[iUrl|VerPath|ResPath] helper functions of 
  //       service/woveon.js. This could lead to versioning problems.

// This should be imported from the secrets file
let testconfig = {
  'user': {
    'id'        : process.env.WOV_plugin_testing_uid, // 3 
    'name'      : process.env.WOV_plugin_testing_uname, // 'tester',
    'email'     : process.env.WOV_plugin_testing_uemail, // tester@woveon.net',
    'password'  : process.env.WOV_plugin_testing_upassword, // 5A...',
    'epassword' : null,                 // encrypted below
  },
  'team': {
    'id'    : process.env.WOV_plugin_testing_tid, // 3,
    'title' : process.env.WOV_plugin_testing_ttitle, // 'Test Team',
    'code'  : process.env.WOV_plugin_testing_tcode, // '5A...',
  },
  'app': {
    'id': process.env.WOV_plugin_id, // 5,
  },
  'teamapp': {'id': process.env.WOV_plugin_testing_taid}, // 5
};
module.exports.testconfig = testconfig;
logger.info('created testconfig : ', testconfig);

module.exports.modelsClear = async () => {
  let db = null;
  try {
    db = new DB(new Logger('helperdb', {debug: true, showName: true}));
    db.logger.info('Helper.modelsClear start');
    let channelModel = null;
    let remoteListenerModel = null;

    if (!db.isConnected()) {
      await db.init(config.plugin.dburl);
      channelModel = ChannelModel(db);
      remoteListenerModel = RemoteListenerModel(db);
      wovMsgsModel = WovMsgsModel(db);
    }
    let chs = await channelModel.remove({});
    chs.result.ok.should.equal(1);
    let ws  = await remoteListenerModel.remove({});
    ws.result.ok.should.equal(1);
    let wmmp = await wovMsgsModel.Posts.remove({});
    wmmp.result.ok.should.equal(1);
    let wmmc = await wovMsgsModel.Comments.remove({});
    wmmc.result.ok.should.equal(1);
    let wmml = await wovMsgsModel.Likes.remove({});
    wmml.result.ok.should.equal(1);

    await db.close();
    db.logger.info('Helper.modelsClear end');
  } catch (err) { db.logger.error(err); throw new Error('Failed in Helper.modelsClear'); }
};


/**
 * Code to encrypt a password with a salt, taken for API code.
 *  - From backend/src/core/models/User.js
 * @param {string} _pw - a raw password
 * @return {string} - hashed and salted password
 */
module.exports.cryptPassword = async (_pw) => {

  return new Promise(
    (resolve, reject) => bcrypt.genSalt(SALT_WORK_FACTOR,
      (error, salt) => {
//        console.log(`salt : '${salt}'`);
        if (error) {return reject(error);}
        return bcrypt.hash(_pw, salt, (err, hash) => {
//          console.log(`hash : '${hash}'`);
          if (err) {return reject(err);}
          return resolve(hash);
        });
      }));
};


// Contact the server and return a session object
module.exports.initAPISession = async () => {



  logger.info(`apiUrl : ${apiUrl}`);

  // console.log(`hashed password : ${testconfig.user.epassword}`);
//  let url = `${apiUrl.protocol}//${apiUrl.hostname}:${apiUrl.port}` +
  logger.info('testconfig: ',testconfig);
  let url = `${apiUrl}/api/v1/user/auth?mail=${testconfig.user.email}&password=${testconfig.user.password}`;
  logger.info('begin create channel in API: ', url);

  return new Promise(async (res,rej) => {
    await requester.get(url)
      .catch((err) => { logger.error(err); process.exit(1); })
      .then((result) => {
        // logger.info('Created Session with API: ', JSON.stringify(result));
        result.success.should.be.true();
        res(result.data);
      })
      .catch((err) => { logger.error('Failed to init session. You probably need to register the test app data and test user: `make register_test-plugin` and/or `make register_test-plugin-accounts`.'); logger.error(err); process.exit(1); });
  });
};

module.exports.getChannel = async (_session) => {
  let url = `${apiUrl.protocol}//${apiUrl.hostname}:${apiUrl.port}/api/v1/user/protect/team/protect/channel`;
  let retval = await requester.get(`${url}?session=${encodeURIComponent(_session.session)}&teamId=${encodeURIComponent(testconfig.team.id)}`);
  return retval;

};


/**
 * Emulates the API's call to your plugin to auth a channel. 
 *   - the user clicks create channel, telling API
 *   - API tells plugin to expect a call
 *   - plugin returns URL and channel token to use {url:, token:}
 *   - API keeps the token, and returns the URL
 *
 * At the end, your plugin will have a placeholder for a channel because /auth was called
 * and this will return a url to your plugin's /remote_auth call.
 * 
 * NOTE: call getChannel to get an existing channel. call this to create and begin authing it.
 * 
 * @param {object} _session
 * @return {url}
 */
module.exports.emulateBrowserChannelAuth = async (_session) => {
  let createChannelOpts = {
    sourceId : testconfig.app.id,
    userId   : _session.id,
    session  : _session.session,
    teamId   : testconfig.team.id,
    title    : testconfig.team.title,
  };
  let url = `${apiUrl.protocol}//${apiUrl.hostname}:${apiUrl.port}/api/v1/user/protect/team/protect/channel`;
  logger.info(`apiUrl: ${apiUrl}`);
  logger.info(`request : ${url} : ${JSON.stringify(createChannelOpts, null, ' ')}`);
  let result = await requester.post(url, null, createChannelOpts);
  console.log(`asked API to create team channel of ${testconfig.team.id}, resulted in: `, result);
  if ( result.success != true ) {
    logger.error(`\n\n` +
      `* ERROR ********************************************************************\n` +
      `* ERROR * Did you run the bin/registerplugintester.sh\n` +
      `* ERROR * script to create the user, team, teamapplication in API's database?\n` +
      `* ERROR *  NOTE: see the API's log for what happened.\n` +
      `* ERROR ********************************************************************\n`);
    throw new Error('Failed Browser Channel Auth.');
  }
  return result;
}

module.exports.old_emulateBrowserChannelAuth = async (_session) => {

  let createChannelOpts = {
    sourceId : testconfig.app.id,
    userId   : _session.id,
    session  : _session.session,
    teamId   : testconfig.team.id,
    title    : testconfig.team.title,
  };

  let url = `${apiUrl.protocol}//${apiUrl.hostname}:${apiUrl.port}/api/v1/user/protect/team/protect/channel`;

  let result = await requester.get(`${url}?session=${encodeURIComponent(_session.session)}&teamId=${encodeURIComponent(testconfig.team.id)}`);
  console.log(`get team channel of ${testconfig.team.id}, is: `, result);

  // return it if found one
  if ( result.success == true ) { return result; }

  // create one if not found
  else { 

    logger.error(`\n\n` +
      `* ERROR ********************************************************************\n` +
      `* ERROR * Did you run the bin/registerplugintester.sh\n` +
      `* ERROR * script to create the user, team, teamapplication in API's database?\n` +
      `* ERROR *  NOTE: see the API's log for what happened.\n` +
      `* ERROR ********************************************************************\n`);
    let result = await requester.post(url, null, createChannelOpts);
    console.log(`created team channel of ${testconfig.team.id}, is: `, result);
    return result;
  }

};


/**
 * A long chain of calls best not looked at too harshly. I have a graphic to display it.
 * @param {Gatekeeper} _gatekeeper
 * @param {object} _session
 * @param {object} _nockRemoteAuth - ex. { url: 'https://graph.facebook.com', method: 'get', result: { access_token: 'AAA' } }
 * @param {object} _nockRemoteAuth2 - ex. { url: 'https://graph.facebook.com', method: 'get', result: { access_token: 'AAA' } }
 * @return {string} - token of the channel newly created and authed
 */
module.exports.createNewChannel = async (_gatekeeper, _session, _nockRemoteAuth= null, _nockRemoteAuth2 = null, _dodelay = null) => {
  let retval = null;
  let fullremoteurl = null;
  let remoteurl= null;
  let newchanneltoken = null;

//  logger.info('get class of gtkeerp ', _gatekeeper.constructor.name);
  if ( _gatekeeper.constructor.name != 'GatekeeperService' ) {
    throw new Error('createNewChannel is only for actual Gatekeeper/Worker objects. Call emulateBrowserChannelAuth and then do custom for Browser Remote Auth');
  }


  if ( _dodelay != null) logger.info('workers 1 : ', await _gatekeeper.RemoteListenerModel.find({}));


  await module.exports.emulateBrowserChannelAuth(_session)
    .then((_result) => {
      _result.success.should.be.true();
      should.notEqual(_result.data, null);

      fullremoteurl = _result.data;
      remoteurl = new URL(_result.data);
      console.log('url parsed: ', remoteurl.toString());
      newchanneltoken = qs.parse(remoteurl.query).t;
      newchanneltoken.should.not.be.null();
      logger.verbose('new authed channel token: ', newchanneltoken);
      retval = newchanneltoken;
    });

//  if ( _dodelay != null) logger.info('workers 2 : ', await _gatekeeper.RemoteListenerModel.find({}));

  // setup nock if you have a predefined remoteAuthResult
  try {
    if (_nockRemoteAuth != null) {
      _nockRemoteAuth.result.access_token = newchanneltoken;
      logger.info('Using nock on :',_nockRemoteAuth);
      let n = nock(_nockRemoteAuth.url)
        .filteringPath(() => { return '/'; })[_nockRemoteAuth.method]('/')
        .reply(200, _nockRemoteAuth.result);
      if (_nockRemoteAuth2 != null) {
      logger.info('Using nock2 on:',_nockRemoteAuth2);
        let n = nock(_nockRemoteAuth2.url)
          .filteringPath(() => { return '/'; })[_nockRemoteAuth2.method]('/')
          .reply(200, _nockRemoteAuth2.result);
      }
    }
  } catch (err) {console.log(err); throw new Error('failed nock in crateNewChannel');};



//  logger.info('workers 3 : ', await _gatekeeper.RemoteListenerModel.find({}));


  if ( _dodelay != null) { console.log("*********************** 10 sec timeout");
    await new Promise((res, rej) => {
      setTimeout(function () { console.log('timout finished'); res(true); }, 10000); }); }

  // Browser Remote Auth
  // _gatekeeper.requester.post(`${remoteurl.address}:${remoteurl.post}/start_channel`, 
  //                             null, {token: newchanneltoken})
  //logger.info('Use gatekeeper\'s toWorkers requester to send to self:', fullremoteurl);
  await _gatekeeper.toWorkers.get(fullremoteurl)
    .catch((err) => { console.log(err); throw new Error('failed _gatekeeper /remote_auth'); })
    .then( async (rresult) => {
      logger.verbose('remote authed channel, returned:', rresult); // 
      rresult.success.should.be.true();

      if (_dodelay != null) {
        console.log("*********************** 10 sec timeout");
        await new Promise((res, rej) => {
          setTimeout(function () { console.log('timout finished'); res(true); }, 10000);
        });
      }

      /*
      Object.keys(rresult).length.should.be.equal(1);
      if ( _nockRemoteAuth != null ) {
        rresult.access_token.should.be.equal(_nockRemoteAuth.result.access_token);
      }
      */

      // building URL to Gatekeeper (NOTE: just using toWorkers requester, does not
      //    actually go to workers here).
      //logger.verbose('Building URL to call /launch on from: ',remoteurl);
      let launchurl = `http://${remoteurl.host}/launch`;
      logger.verbose(` calling ${launchurl}`);
      return _gatekeeper.toWorkers.post(launchurl,null,{token: newchanneltoken });
    })
    .catch((err) => { console.log(err); throw new Error('failed _gatekeeper /launch'); })
    .then((rresult2) => {
      logger.verbose('launch returned, ',rresult2);
      //rresult2.success.should.be.true();
      //rresult2.data.channel.should.equal(retval);
      //should.notEqual(rresult2.data.url,null);
    })
    .catch((err) => { console.log(err); throw new Error('failed _gatekeeper test'); });

//  logger.info('workers 4 : ', await _gatekeeper.RemoteListenerModel.find({}));

  return retval;
};


module.exports.SimulatedRemoteService = class {

  constructor(_curtime,_logger) {
    this.logger = _logger;
    this.engine = null;
    this.posts = [];

    this.resetSim(_curtime);

    autoBind(this);
  }

  linkToEngine(_engine) { this.engine = _engine; this.engine.rs = this; }


  // engine
  async getInc() { return this._curtime; }
  genMsgInc(_rawmsg, _channel) { return _rawmsg != null ? _rawmsg[0]: null; }

  /**
   * 2nd position is good enough for an id.
   * @param {*} _rawmsg
   * @return {string}
   */
  genMsgRemoteID(_rawmsg) { return _rawmsg[1]; }


  /**
   * Reset time, not data.
   * @param {int} _curtime
   * @param {int} _monitor
   */
  resetSim(_curtime,_monitor) {
    this._curtime = _curtime;
    if ( this.engine ) 
      this.engine.resetSim(_monitor);
  }

  /**
   * Go to next time of update, sending update to engine.
   */
  setToTime(_newtime) {
    let bounds = [-1,-1];
    let i = 0;
    this.logger.log(`setToTime`, `setToTime : ${this._curtime} -> ${_newtime}`);

    // find start, save index
    while(i < this.posts.length && this.posts[i][0] <= this._curtime) { i++; }
    bounds[0] = i;

    // go to this _newtime, save index
    while(i < this.posts.length && this.posts[i][0] <= _newtime) { i++; }
    bounds[1] = i;

    this.logger.log(`setToTime`,` - bounds: ${bounds}`);

    // send all updates to engine
    for(i=bounds[0];i<bounds[1];i++) {
      let p = this.posts[i];
      this.engine.addMsg(p[0],p[1]);
    }
    this._curtime = _newtime;
  }


  /**
   * From this time, back, up to limit.
   * @param {*} _time  - start fetching from BELOW here and back (not inclusive)
   * @param {*} _limit  - number to get
   * @return {array} - of posts, in increasing time
   */
  async fetchRSOnAndBefore(_time, _limit) {

    if ( _time == 'now' ) _time = this._curtime;

    let i = -1;
    let bounds = [-1,-1]; // [start,end] where start > end
    let retval = [];

    this.logger.log('fetchRSOnAndBefore', `fetchRSOnAndBefore : t:${_time} l:${_limit}`);

    // find start, starting at end
    i = this.posts.length-1;
    while (i >= 0 && this.posts[i][0] > _time) {i--;} // OnAnd
    bounds[1] = i;

    // find end by counting down and then making sure bigger than 0
    i = i-_limit+1;
    i = i<0? 0:i;
    bounds[0] = i;
    this.logger.log('fetchRSBefore', `  - : ${bounds[0]} <-> ${bounds[1]} : inclusive index to return`);

    // gather those (ordered by default since posts are ordered)
    for (i=bounds[0]; i<=bounds[1]; i++) {
      retval.push( [this.posts[i][0], this.posts[i][1]] );
    }
    this.logger.log('fetchRSBefore', `  - : ${JSON.stringify(retval)}`);

    this.logger.info('returning',retval);
    return retval;
  }


  /**
   * From this time, forward, up to limit.
   * @param {*} _time - fetch from AFTER here and forward (not inclusive)
   * @param {*} _limit - number to get
   */
  async fetchRSOnAndAfter(_time, _limit) {

    if ( _time == 'now' ) _time = this._curtime();

    let i = -1;
    let bounds = [-1, -1]; // [start,end] where start > end
    let retval = [];

    this.logger.info(`fetchRSOnAndAfter: t:${_time} l:${_limit}`);

    // find start, starting at end
    i = this.posts.length;
    let cont = (i > 0?true:false); // ensure always starts with i >= 0
    while ( cont == true ) {
      i--;
      if ( this.posts[i][0] < _time ) {i++; cont = false;}
      else if ( i == 0 ) cont = false;
    }

    // if have data to return
    if (i < this.posts.length) {
      //    if ( i >= this.posts.length) i = this.posts.length-1;
      bounds[0] = i; //(i < this.posts.length?i:this.posts.length-1); // (i< 0? 0: i);

      // find end by counting up and then making sure bigger than 0
      i = i + _limit - 1;
      i = i >= this.posts.length ? this.posts.length - 1 : i;
      bounds[1] = i;
      this.logger.info(`  - : bounding indices: ${bounds[0]} <-> ${bounds[1]}`);

      // gather those (ordered by default since posts are ordered)
      for (i = bounds[0]; i <= bounds[1]; i++) {
        retval.push([this.posts[i][0], this.posts[i][1]]);
      }
    }
    this.logger.info(`  - : ${JSON.stringify(retval)}`);

    return retval;
  }

  /**
   * Adds data.
   * @param {*} _time 
   * @param {*} _data 
   */
  setPost(_time, _data) {
    this.posts.push([_time,_data]);
    this.posts.sort(DownloaderEngine._compareEntries);

  }

  displayMe(_msg) {
    if ( _msg == null ) _msg = '';
    else _msg = `'${_msg}' `;
    this.logger.info(`SimulatedRemoteService: ${_msg}length:${this.posts.length} curtime:${this._curtime}`);
    for(let i=0; i< this.posts.length; i++) {
      let e = this.posts[i];
      this.logger.info(`${i}: '${JSON.stringify(e)}'`);
    }
  }

  compareToEngine() {
    if ( this.posts.length != this.engine.posts.length ) {
      this.logger.error(`Error in ${this.constructor.name}.compareToEngine`);
      this.displayMe();
      this.engine.displayMe();
      throw new Error('Sim and Engine not even same length.');
    } else {
      for(let i=0; i< this.posts.length; i++ ){
        if ( (this.posts[i][0] != this.engine.posts[i][0] ) ||
             (this.posts[i][1] != this.engine.posts[i][1] ) ) {
          this.logger.error(`Error in ${this.constructor.name}.compareToEngine`);
          this.displayMe();
          this.engine.displayMe();
          throw new Error(`Sim and Engine differ at index ${i}`);
        }

      }

    }

  }

};
